# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject, GLib, Gio, Gtk

from time import time

from eolie.define import App
from eolie.utils import emit_signal


class DownloadManager(GObject.GObject):
    """
        Downloads Manager
    """
    __gsignals__ = {
        "download-start": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "download-finish": (GObject.SignalFlags.RUN_FIRST, None, ())
    }

    def __init__(self):
        """
            Init download manager
        """
        GObject.GObject.__init__(self)
        self.__downloads = []
        self.__finished = []

    def add(self, download, filename=None):
        """
            Add a download
            @param download as WebKit2.Download
            @param filename as str/None
        """
        if download not in self.__downloads:
            self.__downloads.append(download)
            download.connect('finished', self.__on_finished)
            download.connect('failed', self.__on_failed)
            download.connect('decide-destination',
                             self.__on_decide_destination, filename)

    def remove(self, download):
        """
            Remove download
            @param download as WebKit2.Download
        """
        if download in self.__downloads:
            self.__downloads.remove(download)
        if download in self.__finished:
            self.__finished.remove(download)

    def get(self):
        """
            Get running downloads
            @return [WebKit2.Download]
        """
        return self.__downloads

    def get_finished(self):
        """
            Get finished download
            @return [WebKit2.Download]
        """
        return self.__finished

    def cancel(self):
        """
            Cancel all downloads
        """
        for download in self.__downloads:
            download.cancel()

    @property
    def active(self):
        """
            Is download active
            @return bool
        """
        for download in self.__downloads:
            if download.get_estimated_progress() != 1.0:
                return True

#######################
# PRIVATE             #
#######################
    def __on_decide_destination(self, download, filename, wanted_filename):
        """
            Modify destination if needed
            @param download as WebKit2.Download
            @param filename as str
            @param wanted_filename as str
        """
        filename = filename.replace("/", "_")
        extension = filename.split(".")[-1]
        if wanted_filename:
            # FIXME We should find a way to pass good extension,
            # fallback to avi
            if extension == filename:
                extension = "avi"
            filename = wanted_filename + "." + extension
        directory_uri = App().settings.get_value('download-uri').get_string()
        if not directory_uri:
            directory = GLib.get_user_special_dir(
                GLib.UserDirectory.DIRECTORY_DOWNLOAD)
            directory_uri = GLib.filename_to_uri(directory, None)
        destination_uri = "%s/%s" % (directory_uri,
                                     GLib.uri_escape_string(filename,
                                                            None,
                                                            False))
        not_ok = True
        i = 1
        try:
            while not_ok:
                f = Gio.File.new_for_uri(destination_uri)
                if f.query_exists():
                    extension_less = filename.replace(".%s" % extension, "")
                    new_filename = "%s_%s.%s" % (extension_less, i, extension)
                    destination_uri = "%s/%s" % (directory_uri,
                                                 GLib.uri_escape_string(
                                                     new_filename,
                                                     None,
                                                     False))
                else:
                    not_ok = False
                i += 1
        except:
            # Fallback to be sure
            destination_uri = "%s/@@%s" % (directory_uri,
                                           GLib.uri_escape_string(
                                               filename,
                                               None,
                                               False))

        webkit_uri = GLib.uri_unescape_string(destination_uri, None)
        download.set_destination(webkit_uri)
        emit_signal(self, "download-start", str(download))
        # Notify user about download
        window = App().active_window
        if window is not None:
            window.toolbar.end.show_download(download)

    def __on_finished(self, download):
        """
            @param download as WebKit2.Download
        """
        self.remove(download)
        self.__finished.append(download)
        emit_signal(self, "download-finish")
        if App().settings.get_value("open-downloads"):
            destination = download.get_destination()
            f = Gio.File.new_for_uri(destination)
            if f.query_exists():
                Gtk.show_uri(None, destination, int(time()))

    def __on_failed(self, download, error):
        """
            @param download as WebKit2.Download
            @param error as GLib.Error
        """
        self.remove(download)
        self.__finished.append(download)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk


class Stack(Gtk.Overlay):
    """
        Compatible with Gtk.Stack API
        Used to get snapshot from WebKit by unmapping webview only when needed
    """

    def __init__(self):
        """
            Init overlay
        """
        Gtk.Overlay.__init__(self)
        self.__visible_child = None

    def add(self, webview):
        """
            Add widget to stack
            @param webview as WebView
        """
        webview.connect("snapshot-changed", self.__on_webview_snapshot_changed)
        webview.connect("destroy", self.__on_webview_destroy)
        self.add_overlay(webview)
        self.reorder_overlay(webview, 0)
        if self.__visible_child is None:
            self.__visible_child = webview

    def set_visible_child(self, webview):
        """
            Set visible child
            @param webview as WebView
        """
        if self.__visible_child is not None and\
                self.__visible_child.is_snapshot_valid:
            self.__visible_child.hide()
        webview.show()
        self.__visible_child = webview
        self.reorder_overlay(webview, -1)

    def get_visible_child(self):
        """
            Get visible child
            @return Gtk.Widget
        """
        return self.__visible_child

#######################
# PRIVATE             #
#######################
    def __on_webview_destroy(self, webview):
        """
            Reset visible child
            @param webview as WebView
        """
        if self.__visible_child == webview:
            self.__visible_child = None

    def __on_webview_snapshot_changed(self, webview, surface):
        """
            We can now hide this view if not visible one
            @param webview as WebView
            @param surface as cairo.Surface
        """
        if webview != self.__visible_child:
            webview.hide()



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from eolie.widget_find import FindWidget
from eolie.widget_uri_label import UriLabelWidget


class OverlayContainer:
    """
        Overlay management for container
    """

    def __init__(self):
        """
            Init container
        """
        grid = Gtk.Grid.new()
        grid.show()
        grid.set_orientation(Gtk.Orientation.VERTICAL)
        self.__find_widget = FindWidget(self._window)
        self.__find_widget.show()
        self._uri_label = UriLabelWidget()
        self.__overlay = Gtk.Overlay.new()
        self.__overlay.show()
        self.__overlay.add_overlay(self._uri_label)
        grid.add(self.__find_widget)
        grid.add(self.__overlay)
        self.add2(grid)

    @property
    def overlay(self):
        """
            Get overlay
            @return Gtk.Overlay
        """
        return self.__overlay

    @property
    def find_widget(self):
        """
            Get find widget
            @return FindWidget
        """
        return self.__find_widget

#######################
# PRIVATE             #
#######################



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject


class Item(GObject.GObject):
    id = GObject.Property(type=int,
                          default=0)
    type = GObject.Property(type=int,
                            default=0)
    title = GObject.Property(type=str,
                             default="")
    uri = GObject.Property(type=str,
                           default="")
    atime = GObject.Property(type=int,
                             default=0)

    def __init__(self):
        GObject.GObject.__init__(self)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk

from eolie.popover_uri_input import Input
from eolie.define import Type


class UriPopoverEvents:
    """
        Events handler for UriPopover
    """

    def __init__(self):
        """
            Init handler
        """
        pass

    def forward_event(self, keyval, state):
        """
            Forward event, smart navigation between boxes
            @param keyval as bool
            @param state as state as Gdk.ModifierType
            @return True if event forwarded
        """
        if not self.is_visible():
            return False
        if keyval == Gdk.KEY_Up and self._input == Input.NONE:
            return False
        elif keyval == Gdk.KEY_Left and self._input == Input.BOOKMARKS:
            self._input = Input.TAGS
            self._tags_box.get_style_context().add_class("kbd-input")
            self._bookmarks_box.get_style_context().remove_class("kbd-input")
            self._window.toolbar.title.entry.set_text_entry("")
            return True
        elif keyval == Gdk.KEY_Right and self._input == Input.TAGS:
            self._input = Input.BOOKMARKS
            self._bookmarks_box.get_style_context().add_class("kbd-input")
            self._tags_box.get_style_context().remove_class("kbd-input")
            return True
        elif keyval in [Gdk.KEY_Left, Gdk.KEY_Right] and\
                self._input == Input.SEARCH:
            return False
        elif keyval in [Gdk.KEY_Left, Gdk.KEY_Right] and\
                self._input != Input.NONE:
            return True
        elif keyval in [Gdk.KEY_Down, Gdk.KEY_Up]:
            # If nothing selected, detect default widget
            if self._input == Input.NONE:
                if self._stack.get_visible_child_name() == "search":
                    self._input = Input.SEARCH
                elif self._stack.get_visible_child_name() == "bookmarks":
                    self._tags_box.get_style_context().add_class("kbd-input")
                    self._input = Input.TAGS
                else:
                    self._input = Input.NONE
                box = self._get_current_input_box()
                if box is not None:
                    rows = box.get_children()
                    if rows:
                        box.select_row(None)
                        box.select_row(rows[0])
                return True
            box = self._get_current_input_box()
            rows = box.get_children()
            if box is None or not rows:
                self._input = Input.NONE
                return False
            selected = box.get_selected_row()
            # If nothing selected, select first row
            if selected is None:
                box.select_row(rows[0])
                if self._input == Input.TAGS:
                    item_id = rows[0].item.get_property("id")
                    self._set_bookmarks(item_id)
                return True
            else:
                idx = -1 if keyval == Gdk.KEY_Up else 1
                for row in rows:
                    if row == selected:
                        break
                    idx += 1
                box.select_row(None)
                if idx >= len(rows):
                    box.select_row(rows[0])
                    if self._input == Input.TAGS:
                        item_id = rows[0].item.get_property("id")
                        self._set_bookmarks(item_id)
                    return True
                elif idx < 0:
                    # Do not go to uribar for bookmarks & search
                    if self._input in [Input.BOOKMARKS, Input.SEARCH]:
                        box.select_row(rows[-1])
                        return True
                    else:
                        box.select_row(None)
                        self._input = Input.NONE
                        box.get_style_context().remove_class("kbd-input")
                else:
                    box.select_row(rows[idx])
                    if self._input == Input.TAGS:
                        item_id = rows[idx].item.get_property("id")
                        self._set_bookmarks(item_id)
                    return True
        elif keyval in [Gdk.KEY_Return, Gdk.KEY_KP_Enter]:
            box = self._get_current_input_box()
            if box is not None:
                selected = box.get_selected_row()
                if selected is not None:
                    container = self._window.container
                    item = selected.item
                    # Switch to view
                    if item.get_property("type") == Type.WEBVIEW:
                        title = item.get_property("title")
                        uri = item.get_property("uri")
                        for webview in container.webviews:
                            if webview.uri == uri and webview.title == title:
                                container.set_visible_webview(webview)
                                self._window.close_popovers()
                                break
                    # Load URI
                    else:
                        uri = selected.item.get_property("uri")
                        if uri:
                            self._window.close_popovers()
                            container.webview.load_uri(uri)
                            container.set_expose(False)
                            return True
            else:
                self._input = Input.NONE
        else:
            self._input = Input.NONE
        return False



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, GtkSpell

from eolie.helper_gestures import GesturesHelper
from eolie.define import App


class LanguageRow(Gtk.ListBoxRow, GesturesHelper):
    """
        Language row (Allowing to select a language for uri)
    """

    def __init__(self, uri, name, code):
        """
            Init row
            @param uri as str
            @param name as str
            @param code as str
        """
        Gtk.EventBox.__init__(self)
        self.__uri = uri
        self.__code = code
        grid = Gtk.Grid()
        grid.set_column_spacing(5)
        grid.show()
        label = Gtk.Label.new(name)
        label.set_hexpand(True)
        label.set_property("halign", Gtk.Align.START)
        label.show()
        self.__check = Gtk.CheckButton()
        self.__check.show()
        grid.add(self.__check)
        grid.add(label)
        eventbox = Gtk.EventBox.new()
        eventbox.show()
        eventbox.add(grid)
        self.add(eventbox)
        GesturesHelper.__init__(self, eventbox)
        user_code = ""
        codes = App().websettings.get_languages(uri)
        if not codes:
            codes = []
            locales = GLib.get_language_names()
            if locales:
                user_code = locales[0].split(".")[0]
                codes = [user_code]
        self.__check.set_active(codes is not None and code in codes)
        self.__check.connect("toggled", self.__on_check_toggled)
        GesturesHelper.__init__(self, eventbox)

#######################
# PROTECTED           #
#######################
    def _on_primary_press_gesture(self, x, y, event):
        """
            Toggle check button
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        toggled = not self.__check.get_active()
        self.__check.set_active(toggled)

#######################
# PRIVATE             #
#######################
    def __on_check_toggled(self, check):
        """
            Save state
            @param check as Gtk.CheckButton
        """
        active = check.get_active()
        if active:
            App().websettings.add_language(self.__code, self.__uri)
        else:
            App().websettings.remove_language(self.__code, self.__uri)
        App().active_window.container.webview.update_spell_checking(
            self.__uri)


class LanguagesMenu(Gtk.Bin):
    """
        Widget showing languages and allowing user to enable/disable
        spell check
    """

    def __init__(self, uri):
        """
            Init widget
            @param uri as str
        """
        Gtk.Bin.__init__(self)
        self.__uri = uri
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/LanguagesMenu.ui")
        builder.connect_signals(self)
        self.__switch = builder.get_object("switch")
        self.add(builder.get_object("languages"))

#######################
# PROTECTED           #
#######################
    def _on_map(self, listbox):
        """
            Populate languages
            @param listbox as Gtk.ListBox
        """
        enable_spell_checking = App().settings.get_value("enable-spell-check")
        self.__switch.set_active(enable_spell_checking)
        if not listbox.get_children():
            checker = GtkSpell.Checker()
            for language in checker.get_language_list():
                name = checker.decode_language_code(language)
                row = LanguageRow(self.__uri, name, language)
                row.show()
                listbox.add(row)

    def _on_state_set(self, listbox, state):
        """
            Save spell checking state
            @param listbox as Gtk.ListBox
            @param state as bool
        """
        App().settings.set_value("enable-spell-check",
                                 GLib.Variant("b", state))
        listbox.set_sensitive(state)
        for window in App().windows:
            for webview in window.container.webviews:
                context = webview.get_context()
                context.set_spell_checking_enabled(state)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Gio

from gettext import gettext as _

from eolie.define import App
from eolie.logger import Logger


class ImportBookmarksDialog:
    """
        Import bookarks dialog
    """

    def __init__(self, window):
        """
            Init widget
            @param window as Gtk.Window
        """
        self.__window = window
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/DialogImportBookmarks.ui")
        builder.connect_signals(self)
        self.__dialog = builder.get_object("dialog")
        self.__dialog.set_transient_for(window)
        self.__listbox = builder.get_object("listbox")
        self.__firefox_items = self.__get_firefox_profiles()
        items = []
        for (name, path) in self.__firefox_items:
            items.append("Firefox: %s" % name)
        items += ["Chromium", "Chrome"]
        try:
            from bs4 import BeautifulSoup
            BeautifulSoup
            items.append(_("Others"))
        except Exception as e:
            Logger.error("ImportBookmarksDialog::__init__(): %s", e)
        for item in items:
            label = Gtk.Label.new(item)
            label.show()
            self.__listbox.add(label)
        headerbar = builder.get_object("headerbar")
        self.__dialog.set_titlebar(headerbar)

    def run(self):
        """
            Run dialog
        """
        self.__dialog.run()
        self.__dialog.destroy()

#######################
# PROTECTED           #
#######################
    def _on_dialog_response(self, dialog, response_id):
        """
            Import bookmarks
            @param dialog as Gtk.Dialog
            @param response_id as int
        """
        if response_id == Gtk.ResponseType.DELETE_EVENT:
            return
        label = self.__listbox.get_selected_row().get_child().get_label()
        if label.startswith("Firefox: "):
            profile = label.replace("Firefox: ", "")
            for item in self.__firefox_items:
                if item[0] == profile:
                    App().bookmarks.import_firefox(item[1])
                    break
        elif label == "Chrome":
            App().bookmarks.import_chromium(True)
        elif label == "Chromium":
            App().bookmarks.import_chromium(False)
        else:
            dialog = Gtk.FileChooserNative.new(
                _("Import HTML bookmarks"), self.__window,
                Gtk.FileChooserAction.OPEN,
                _("Import"),
                _("Cancel"))
            dialog.connect("response", self.__on_file_chooser_response)
            dialog.run()
            dialog.destroy()
        self.__window.close_popovers()

#######################
# PRIVATE             #
#######################
    def __get_firefox_profiles(self):
        """
            Get available firefox profiles
            @return [(str, str)]
        """
        profiles = []
        try:
            firefox_path = GLib.get_home_dir() +\
                "/.mozilla/firefox/profiles.ini"
            f = Gio.File.new_for_path(firefox_path)
            (status, contents, tags) = f.load_contents()
            data = contents.decode("utf-8")
            name = path = None
            for item in data.split("\n"):
                if item.startswith("Name="):
                    name = item.split("=")[1]
                if item.startswith("Path="):
                    path = item.split("=")[1]
                if path is not None and name is not None:
                    profiles.append((name, path))
                    name = path = None
        except Exception as e:
            Logger.error(
                "ImportBookmarksDialog::__get_firefox_profiles(): %s", e)
        return profiles

    def __on_file_chooser_response(self, dialog, response_id):
        """
            Import file
            @param dialog as GtkFileChooserDialog
            @param response_id as int
        """
        if response_id == Gtk.ResponseType.ACCEPT:
            path = dialog.get_filename()
            App().bookmarks.import_html(path)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk, Gio, WebKit2, GObject

from gettext import gettext as _
from urllib.parse import urlparse
from time import time

from eolie.define import App, LoadingState
from eolie.utils import emit_signal
from eolie.webview_signals_menu import WebViewMenuSignals
from eolie.webview_signals_js import WebViewJsSignals


class WebViewSignals(WebViewMenuSignals, WebViewJsSignals):
    """
        Handle webview signals
    """

    gsignals = {
        "is-playing-audio": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
        "readability-content": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "readability-status": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
        "title-changed": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "uri-changed": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "snapshot-changed": (GObject.SignalFlags.RUN_FIRST, None,
                             (GObject.TYPE_PYOBJECT,)),
    }

    for signal in gsignals:
        args = gsignals[signal]
        GObject.signal_new(signal, WebKit2.WebView,
                           args[0], args[1], args[2])

    def __init__(self):
        """
            Init handler
            @param webview as WebView
        """
        WebViewMenuSignals.__init__(self)
        WebViewJsSignals.__init__(self)
        self.reset_last_click_event()
        self.__cancellable = Gio.Cancellable()
        self.__uri = ""
        self.__title = ""
        self.connect("notify::title", self.__on_title_changed)
        self.connect("notify::uri", self.__on_uri_changed)
        self.connect("scroll-event", self.__on_scroll_event)
        self.connect("run-file-chooser", self.__on_run_file_chooser)
        self.connect("button-press-event", self.__on_button_press_event)

    def reset_last_click_event(self):
        """
            Reset last click event
        """
        self._last_click_event_x = 0
        self._last_click_event_y = 0
        self._last_click_time = 0

    def set_title(self, title):
        """
            Set title
            @param title as str
        """
        if title:
            self.__title = title
            emit_signal(self, "title-changed", title)

    def set_uri(self, uri):
        """
            Set delayed uri
            @param uri as str
        """
        self.__uri = uri.rstrip("/")
        emit_signal(self, "uri-changed", uri)

    @property
    def uri(self):
        """
            Get webview uri
            @return str
        """
        return self.__uri

    @property
    def title(self):
        """
            Get webview title
            @return str
        """
        return self.__title

#######################
# PRIVATE             #
#######################
    def __on_button_press_event(self, webview, event):
        """
            Hide Titlebar popover
            @param webview as WebView
            @param event as Gdk.Event
        """
        self._last_click_event_x = event.x
        self._last_click_event_y = event.y
        self._last_click_time = time()
        if event.button == 8:
            self.go_back()
            return True
        elif event.button == 9:
            self.go_forward()
            return True
        elif self.get_ancestor(Gtk.Popover) is None:
            return self.window.close_popovers()

    def __on_run_file_chooser(self, webview, request):
        """
            Run own file chooser
            @param webview as WebView
            @param request as WebKit2.FileChooserRequest
        """
        dialog = Gtk.FileChooserNative.new(_("Select files to upload"),
                                           self.window,
                                           Gtk.FileChooserAction.OPEN,
                                           _("Open"),
                                           _("Cancel"))
        dialog.set_select_multiple(request.get_select_multiple())
        chooser_uri = App().websettings.get("chooser_uri", webview.uri)
        if chooser_uri is not None:
            dialog.set_current_folder_uri(chooser_uri)
        response = dialog.run()
        if response in [Gtk.ResponseType.DELETE_EVENT,
                        Gtk.ResponseType.CANCEL]:
            request.cancel()
        else:
            request.select_files(dialog.get_filenames())
            App().websettings.set("chooser_uri",
                                  webview.uri,
                                  dialog.get_current_folder_uri())
        return True

    def __on_scroll_event(self, webview, event):
        """
            Adapt scroll speed to device
            @param webview as WebView
            @param event as Gdk.EventScroll
        """
        source = event.get_source_device().get_source()
        if event.state & Gdk.ModifierType.CONTROL_MASK:
            if source == Gdk.InputSource.MOUSE:
                if event.delta_y < 0.5:
                    webview.zoom_in()
                elif event.delta_y > 0.5:
                    webview.zoom_out()
            else:
                if event.delta_y > 0.5:
                    webview.zoom_in()
                elif event.delta_y < - 0.5:
                    webview.zoom_out()
            return True
        elif source == Gdk.InputSource.MOUSE:
            event.delta_x *= 2
            event.delta_y *= 2

    def __on_uri_changed(self, webview, param):
        """
            Handle JS updates
            @param webview as WebKit2.WebView
            @param param as GObject.ParamSpec
        """
        uri = webview.get_property(param.name).rstrip("/")
        if not uri.startswith("javascript:"):
            self.__uri = uri
            emit_signal(self, "uri-changed", uri)
            self._set_user_agent(uri)

    def __on_title_changed(self, webview, param):
        """
            We launch Readability.js at page loading finished.
            @param webview as WebKit2.WebView
            @param param as GObject.ParamSpec
        """
        title = webview.get_property(param.name)
        if title:
            parsed = urlparse(webview.uri)
            is_http = parsed.scheme in ["http", "https"]
            if is_http:
                self.__title = "%s: %s" % (parsed.netloc, title)
                emit_signal(self, "title-changed", self.__title)
            else:
                self.__title = title
                emit_signal(self, "title-changed", None)
            if self.loading_state in [LoadingState.ERROR,
                                      LoadingState.STOPPED] or\
                    webview.is_ephemeral or\
                    not is_http:
                return
            mtime = round(time(), 2)
            history_id = App().history.add(self.__title, self.__uri, mtime)
            App().history.set_page_state(self.__uri, mtime)
            if App().sync_worker is not None:
                App().sync_worker.push_history(history_id)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Pango, GLib


class UriLabelWidget(Gtk.EventBox):
    """
        Small label trying to not be under mouse pointer
    """

    def __init__(self):
        """
            Init label
        """
        Gtk.EventBox.__init__(self)
        self.__label = Gtk.Label.new()
        self.__label.set_ellipsize(Pango.EllipsizeMode.END)
        self.__label.get_style_context().add_class("urilabel")
        self.__label.show()
        self.add(self.__label)
        self.connect("enter-notify-event", self.__on_enter_notify)

    def set_text(self, text):
        """
            Set label text
            @param text as str
        """
        if text == self.__label.get_text():
            return
        self.set_property("halign", Gtk.Align.END)
        self.set_property("valign", Gtk.Align.END)
        self.__label.get_style_context().add_class("bottom-right")
        self.__label.get_style_context().remove_class("bottom-left")
        self.__label.get_style_context().remove_class("top-left")
        self.__label.set_text(text)

#######################
# PRIVATE             #
#######################
    def __on_enter_notify(self, widget, event):
        """
            Try to go away from mouse cursor
            @param widget as Gtk.Widget
            @param event as Gdk.Event
        """
        GLib.idle_add(self.hide)
        # Move label at the right
        if self.get_property("halign") == Gtk.Align.END:
            self.set_property("halign", Gtk.Align.START)
            self.__label.get_style_context().add_class("bottom-left")
            self.__label.get_style_context().remove_class("bottom-right")
            self.__label.get_style_context().remove_class("top-left")
        # Move label at top
        else:
            self.set_property("halign", Gtk.Align.END)
            self.set_property("valign", Gtk.Align.END)
            self.__label.get_style_context().add_class("top-left")
            self.__label.get_style_context().remove_class("bottom-left")
            self.__label.get_style_context().remove_class("bottom-right")
        GLib.idle_add(self.show)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from urllib.parse import urlparse
import re
from os.path import dirname

from eolie.logger import Logger


class CSSImportRule:
    """
        Represent an import rule
    """

    def __init__(self, css, uri, cancellable):
        """
            Init rule
            @param css as str
            @param uri as str
            @param cancellable as Gio.Cancellable
        """
        self.__stylesheet = None
        try:
            parsed = urlparse(uri)
            search = re.search(r'@import url\(["\']?([^"\')]*)', css)
            css = search.group(1)
            if css.startswith(".."):
                path_split = parsed.path.split("/")
                css_uri = "%s://%s%s/%s" % (parsed.scheme, parsed.netloc,
                                            "/".join(path_split[:-1]), css)
            elif css.startswith("//"):
                css_uri = "%s:%s" % (parsed.scheme, css)
            elif not css.startswith("http"):
                parent = dirname(parsed.path)
                css_uri = "%s://%s%s/%s" % (
                    parsed.scheme, parsed.netloc, parent, css)
            from eolie.css_stylesheet import StyleSheet
            self.__stylesheet = StyleSheet(uri=css_uri,
                                           cancellable=cancellable)
            self.__stylesheet.populate()
        except Exception as e:
            Logger.error("CSSImportRule::__init__: %s -> %s", e, css)

    @property
    def css_text(self):
        """
            Get css text for rules
            @return str
        """
        if self.__stylesheet is None:
            return ""
        else:
            return self.__stylesheet.css_text

    @property
    def populated(self):
        """
            True if rule is populated
            @return bool
        """
        return self.__stylesheet is not None and self.__stylesheet.populated

#######################
# PRIVATE             #
#######################



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
from colorsys import rgb_to_hls

from eolie.logger import Logger
from eolie.define import COLORS


class CSSStyleRule:
    """
        Represent a style rule
    """

    def __init__(self, css):
        """
            Init rule
            @param css as str
        """
        self.__selector = ""
        self.__variables = []
        self.__color_str = None
        self.__background_color_str = None
        self.__background_image_str = None
        self.__background_str = None
        self.__border_str = None
        self.__has_background_url = False
        # Get selectors
        search = re.search('^([^{]*){(.*)}', css)
        if search is not None:
            self.__selector = search.group(1).strip()
            declarations = search.group(2)
            self.__has_background_url = declarations.find("url(") != -1
            for declaration in declarations.split(";"):
                if declaration.find(":") == -1:
                    continue
                (prop, value) = declaration.split(":", 1)
                if not self.__contains_color(value):
                    continue
                prop = prop.strip()
                # This is a variable
                if prop.startswith("--") and value.find("var(") == -1:
                    self.__variables.append((prop, value))
                elif prop == "color":
                    self.__color_str = self.__get_clean_value(value)
                elif prop == "background-color":
                    self.__background_color_str = self.__get_clean_value(value)
                elif prop == "background-image":
                    self.__background_image_str = self.__get_clean_value(value)
                elif prop == "background":
                    self.__background_str = self.__get_clean_value(value)
                elif prop.startswith("border"):
                    self.__border_str = self.__get_clean_value(value)
            if self.__color_str is not None:
                self.__update_color()
            if self.__background_color_str is not None:
                self.__update_background_color()
            if self.__background_image_str is not None:
                self.__update_background_image()
            if self.__background_str is not None:
                self.__update_background()
            if self.__border_str is not None:
                self.__update_border_color()
            if self.__variables:
                self.__update_variables_color()

    @property
    def css_text(self):
        """
            Get css text for rules
            @return str
        """
        rules = []
        if self.__color_str is not None:
            rules.append("color: %s" % self.__color_str)
        if self.__background_color_str is not None:
            rules.append("background-color: %s" % self.__background_color_str)
        if self.__background_image_str is not None:
            rules.append("background-image: %s" % self.__background_image_str)
        if self.__background_str is not None:
            rules.append("background: %s" % self.__background_str)
        if self.__border_str is not None:
            rules.append("border-color: %s" % self.__border_str)
        for (prop, value) in self.__variables:
            rules.append("%s : %s" % (prop, value))
        if rules:
            css_text = "%s{ %s } " % (self.__selector, ";".join(rules))
            if self.__has_background_url:
                css_text += """%s > *
                    {text-shadow: 0px 0px 5px black
                    !important}""" % self.__selector
            return css_text
        else:
            return ""

    @property
    def populated(self):
        """
            True if rule is populated
            @return bool
        """
        return True

#######################
# PRIVATE             #
#######################
    def __get_clean_value(self, value):
        """
            Remove unwanted parts from value
            @param value as str
            @return str
        """
        value = re.sub('[ ]*!.*important', '', value)
        value = re.sub(r'url.*\([^\)]*\)', 'url()', value)
        return value.strip()

    def __contains_color(self, value):
        """
            True if value contains a color
            @param value as str
            @return bool
        """
        ignores = ["inherit", "transparent", "data:",
                   "unset", "currentcolor"]
        for ignore in ignores:
            if value.find(ignore) != -1:
                return False
        # Check hsl/rgb/# colors
        if value.find("hsl") != -1 or\
                value.find("rgb") != -1 or\
                value.find("#") != -1:
            return True
        # Check named colors
        else:
            for color in COLORS.keys():
                if value.find(color) != -1:
                    return True
        return False

    def __get_hsla_as_float(self, value):
        """
            Convert percent str to float, keep value if no percent
            @param value as str
            @return float
        """
        if value.find("%") != -1:
            return float(value[:-1]) / 100
        return float(value)

    def __get_rgba_as_int(self, value):
        """
            Convert percent str to int, keep value if no percent
            @param value as str
            @return int
        """
        if value.find("%") != -1:
            return int(value[:-1]) * 255 / 100
        return int(value)

    def __get_deg(self, deg):
        """
            Convert deg str to int
            @param deg as str
            @return int
        """
        if deg[0].find("deg") != -1:
            return deg[0][:-3]
        return deg

    def __get_split(self, line):
        """
            Split RGBA/HSLA notation (CSS3 vs CSS4 syntax)
        """
        if "," in line:
            return line.split(",")
        else:
            # Get Opacity after /
            slash_split = line.split("/")
            # Get HSL/RGB before /
            split = slash_split[0].split(" ")
            if len(slash_split) == 1:
                # HSL/RGB + no opacity
                return split + [1]
            else:
                # HSL/RGB + A
                return split + slash_split[1]

    def __get_hex_colors_from_rule(self, rule):
        """
            Get RGBA colors as hexa from rule
            @param rule as str
            @return {"hex_color": (int, int, int, int)}
        """
        results = {}
        # Extract values from hexadecimal notation
        colors = re.findall(
            '(#[0-9A-Fa-f]{8}|#[0-9A-Fa-f]{6}|#[0-9A-Fa-f]{3})', rule)
        for color in colors:
            hexa = color[1:]
            a = 1
            # Translate short version of hexa string
            if len(hexa) == 3:
                _hexa = ""
                for letter in hexa:
                    _hexa += letter + letter
                hexa = _hexa
            rgb = tuple(int(hexa[i:i+2], 16) for i in (0, 2, 4))
            if len(hexa) == 8:
                a = round(255 / int(hexa[6, 8], 16))
            results[color] = (rgb + (a,))
        return results

    def __get_rgba_colors_from_rule(self, rule):
        """
            Get RGBA colors from rule
            @param rule str
            @return {"rgba_color": (int, int, int, int)}
        """
        results = {}
        # Extract values from rgb() and rgba()
        colors = re.findall(r'(rgb.?\([^\)]*\))', rule)
        for color in colors:
            color_tuple = re.search(r'rgb.?\(([^\)]*)', color)
            if color_tuple is None:
                continue
            split = self.__get_split(color_tuple[1])
            r = self.__get_rgba_as_int(split[0])
            g = self.__get_rgba_as_int(split[1])
            b = self.__get_rgba_as_int(split[2])
            # Indentical to hsla float calculation
            a = 1
            if len(split) == 4:
                a = self.__get_hsla_as_float(split[3])
            results[color] = (r, g, b, a)
        return results

    def __get_hsla_colors_from_rule(self, rule):
        """
            Get HSLA colors from rule
            @return {"hsla_color": (int, float, float, float)}
        """
        results = {}
        # Extract values from hsl() and hsla()
        colors = re.findall(r'(hsl.?\([^\)]*\))', rule)
        for color in colors:
            color_tuple = re.search(r'hsl.?\(([^\)]*)', color)
            if color_tuple is None:
                continue
            split = self.__get_split(color_tuple[1])
            h = self.__get_deg(split[0])
            s = self.__get_hsla_as_float(split[1])
            l = self.__get_hsla_as_float(split[2])
            a = 1
            if len(split) == 4:
                a = self.__get_hsla_as_float(split[3])
            results[color] = (h, s, l, a)
        return results

    def __get_colors_from_rule(self, rule):
        """
            Get RGBA colors from rule
            @param rule as str
            @return [(h, s, l, a)] as [(int, float, float, float)]
        """
        hex_colors = self.__get_hex_colors_from_rule(rule)
        rgba_colors = self.__get_rgba_colors_from_rule(rule)
        named_colors = {}
        for color in COLORS.keys():
            if rule.find(color) != -1:
                named_colors[color] = COLORS[color] + (1,)
        colors = {**hex_colors, **rgba_colors, **named_colors}
        for key in colors.keys():
            rgba = colors[key]
            (h, l, s) = rgb_to_hls(rgba[0] / 255,
                                   rgba[1] / 255,
                                   rgba[2] / 255)
            colors[key] = (h, s, l, rgba[3])
        hsla_colors = self.__get_hsla_colors_from_rule(rule)
        return {**colors, **hsla_colors}

    def __get_hsla_to_css_string(self, hsla):
        """
            Convert hsla values to CSS string
            @param hsla as (int, float, float, float)
            return str
        """
        return "hsla({}, {}%, {}%, {})".format(hsla[0] * 360,
                                               hsla[1] * 100,
                                               hsla[2] * 100,
                                               hsla[3])

    def __update_background_color(self):
        """
            Update background color for night mode
        """
        try:
            colors = self.__get_colors_from_rule(self.__background_color_str)
            if not colors:
                return None
            for key in colors.keys():
                hsla = colors[key]
                if hsla[3] < 0.4:
                    continue
                elif hsla[1] > 0.2:
                    if hsla[2] > 0.7:
                        hsla = (hsla[0], hsla[1], 0.1, hsla[3])
                    else:
                        hsla = (hsla[0], hsla[1], 0.3, hsla[3])
                else:
                    hsla = (0, 0, 0.21, 1)
                hsla_str = self.__get_hsla_to_css_string(hsla)
                self.__background_color_str =\
                    self.__background_color_str.replace(
                        key, hsla_str)
            self.__background_color_str += " !important;"
        except Exception as e:
            Logger.warning("CSSStyleRule::__update_background_color(): %s", e)
            self.__background_color_str = "#353535 !important;"

    def __update_background_image(self):
        """
            Update background image for night mode
        """
        try:
            colors = self.__get_colors_from_rule(self.__background_image_str)
            if not colors:
                return None
            for key in colors.keys():
                hsla = colors[key]
                if hsla[3] < 0.4:
                    continue
                elif hsla[1] > 0.2:
                    if hsla[2] > 0.7:
                        hsla = (hsla[0], hsla[1], 0.1, hsla[3])
                    else:
                        hsla = (hsla[0], hsla[1], 0.3, hsla[3])
                else:
                    hsla = (0, 0, 0.21, 1)
                hsla_str = self.__get_hsla_to_css_string(hsla)
                self.__background_image_str =\
                    self.__background_image_str.replace(
                        key, hsla_str)
            self.__background_image_str += " !important;"
        except Exception as e:
            Logger.warning("CSSStyleRule::__update_background_image(): %s", e)
            self.__background_image_str = "#353535 !important;"

    def __update_background(self):
        """
            Update background for night mode
        """
        try:
            colors = self.__get_colors_from_rule(self.__background_str)
            if not colors:
                return None
            for key in colors.keys():
                hsla = colors[key]
                if hsla[3] < 0.4:
                    continue
                hsla = (0, 0, 0.21, 1)
                hsla_str = self.__get_hsla_to_css_string(hsla)
                self.__background_str =\
                    self.__background_str.replace(
                        key, hsla_str)
            self.__background_str += " !important;"
        except Exception as e:
            Logger.warning("CSSStyleRule::__update_background(): %s", e)
            self.__background_str = "#353535 !important;"

    def __update_color(self):
        """
            Update color for night mode
        """
        try:
            colors = self.__get_colors_from_rule(self.__color_str)
            if not colors:
                return None
            for key in colors.keys():
                hsla = colors[key]
                hsla_str = self.__get_hsla_to_css_string(
                    (hsla[0], hsla[1], 0.8, hsla[3]))
                self.__color_str =\
                    self.__color_str.replace(
                        key, hsla_str)
            self.__color_str += " !important;"
        except Exception as e:
            Logger.warning("CSSStyleRule::__update_color(): %s", e)
            self.__color_str = "#EAEAEA !important;"

    def __update_border_color(self):
        """
            Update border color for night mode
        """
        try:
            colors = self.__get_colors_from_rule(self.__border_str)
            if not colors:
                return None
            for key in colors.keys():
                hsla = colors[key]
                hsla_str = self.__get_hsla_to_css_string(
                        (hsla[0], hsla[1], 0.3, hsla[3]))
                self.__border_str =\
                    self.__border_str.replace(
                        key, hsla_str)
            self.__border_str += " !important;"
        except Exception as e:
            Logger.warning("CSSStyleRule::__update_border_color(): %s", e)
            self.__border_str = "#EAEAEA !important;"

    def __update_variables_color(self):
        """
            Update variables color for night mode
        """
        variables = []
        for (prop, value) in self.__variables:
            try:
                colors = self.__get_colors_from_rule(value)
                if not colors:
                    continue
                for key in colors.keys():
                    hsla = colors[key]
                    if hsla[2] > 0.75:
                        hsla_str = self.__get_hsla_to_css_string(
                                (0, 0, 0.21, hsla[3]))
                    elif hsla[2] > 0.5:
                        hsla_str = self.__get_hsla_to_css_string(
                            (0, 0, 1 - hsla[2], hsla[3]))
                    else:
                        hsla_str = self.__get_hsla_to_css_string(
                            (hsla[0], hsla[1], 0.5 + hsla[2], hsla[3]))
                    value = value.replace(key, hsla_str)
            except Exception as e:
                Logger.warning(
                    "CSSStyleRule::__update_variables_color(): %s", e)
            variables.append((prop, "%s !important;" % value))
        self.__variables = variables



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib

from urllib.parse import urlparse

from eolie.define import Type, App
from eolie.popover_uri_item import Item
from eolie.popover_uri_row import Row


class UriPopoverSuggestions:
    """
        Suggestions handler for UriPopover
    """

    def __init__(self):
        """
            Init handler
        """
        self.__suggestion_id = None

    def add_suggestions(self, value, cancellable):
        """
            Add suggestions for value
            @param value as str
            @param cancellable as Gio.Cancellable
        """
        parsed = urlparse(value)
        is_uri = parsed.scheme in ["about, http", "file", "https", "populars"]
        # Remove any pending suggestion search
        if self.__suggestion_id is not None:
            GLib.source_remove(self.__suggestion_id)
            self.__suggestion_id = None
        # Search for suggestions if needed
        if App().settings.get_value("enable-suggestions") and not is_uri:
            self.__suggestion_id = GLib.timeout_add(
                100,
                self.__on_suggestion_timeout,
                value,
                cancellable)

#######################
# PRIVATE             #
#######################
    def __add_suggestions(self, suggestions):
        """
            Add suggestions to popover
            @param suggestions as [str]
        """
        for suggestion in suggestions[:2]:
            if suggestion:
                item = Item()
                item.set_property("type", Type.SUGGESTION)
                item.set_property("title", suggestion)
                item.set_property("uri",
                                  App().search.get_search_uri(suggestion))
                child = Row(item, self._window)
                child.show()
                self._search_box.insert(child, 0)

    def __on_search_suggestion(self, uri, status, content, encoding, value):
        """
            Add suggestions
            @param uri as str
            @param status as bool
            @param content as bytes
            @param encoding as str
            @param value as str
        """
        if status:
            string = content.decode(encoding)
            # format: '["{"words"}",["result1","result2"]]'
            sgs = string.replace('[', '')\
                        .replace(']', '')\
                        .replace('"', '')\
                        .split(',')[1:]
            self.__add_suggestions(sgs)

    def __on_suggestion_timeout(self, value, cancellable):
        """
            Search suggestions
            @param value as str
            @param cancellable as Gio.Cancellable
        """
        self.__suggestion_id = None
        App().search.search_suggestions(value,
                                        cancellable,
                                        self.__on_search_suggestion)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gio

import sqlite3
import itertools
from urllib.parse import urlparse
from threading import Lock

from eolie.utils import noaccents, get_random_string
from eolie.define import EOLIE_DATA_PATH, Type
from eolie.localized import LocalizedCollation
from eolie.sqlcursor import SqlCursor
from eolie.logger import Logger
from eolie.database_upgrade import DatabaseUpgrade


class DatabaseBookmarks:
    """
        Eolie bookmarks db
    """

    DB_PATH = "%s/bookmarks.db" % EOLIE_DATA_PATH

    # SQLite documentation:
    # In SQLite, a column with type INTEGER PRIMARY KEY
    # is an alias for the ROWID.
    # Here, we define an id INT PRIMARY KEY but never feed it,
    # this make VACUUM not destroy rowids...
    __create_bookmarks = '''CREATE TABLE bookmarks (
                                               id INTEGER PRIMARY KEY,
                                               title TEXT NOT NULL,
                                               uri TEXT NOT NULL,
                                               popularity INT NOT NULL,
                                               atime REAL NOT NULL,
                                               guid TEXT NOT NULL,
                                               mtime REAL NOT NULL,
                                               position INT DEFAULT 0,
                                               startup INT DEFAULT 0
                                               )'''
    __create_tags = '''CREATE TABLE tags (id INTEGER PRIMARY KEY,
                                          title TEXT NOT NULL)'''
    __create_bookmarks_tags = '''CREATE TABLE bookmarks_tags (
                                                    id INTEGER PRIMARY KEY,
                                                    bookmark_id INT NOT NULL,
                                                    tag_id INT NOT NULL)'''
    # Only useful for Firefox compatibility
    __create_parents = '''CREATE TABLE parents (
                                        id INTEGER PRIMARY KEY,
                                        bookmark_id INT NOT NULL,
                                        parent_guid TEXT NOT NULL,
                                        parent_name TEXT NOT NULL)'''

    def __init__(self):
        """
            Create database tables or manage update if needed
        """
        upgrade = DatabaseUpgrade(Type.BOOKMARK)
        self.thread_lock = Lock()
        if not GLib.file_test(self.DB_PATH, GLib.FileTest.IS_REGULAR):
            try:
                if not GLib.file_test(EOLIE_DATA_PATH, GLib.FileTest.IS_DIR):
                    GLib.mkdir_with_parents(EOLIE_DATA_PATH, 0o0750)
                # Create db schema
                with SqlCursor(self, True) as sql:
                    sql.execute(self.__create_bookmarks)
                    sql.execute(self.__create_tags)
                    sql.execute(self.__create_bookmarks_tags)
                    sql.execute(self.__create_parents)
                    sql.execute("PRAGMA user_version=%s" % upgrade.version)
            except Exception as e:
                Logger.error("DatabaseBookmarks::__init__(): %s", e)
        else:
            upgrade.upgrade(self)

    def add(self, title, uri, guid, tags, atime=0):
        """
            Add a new bookmark
            @param title as str
            @param uri as str
            @param guid as str
            @param tags as [str]
            @param parent_guid as str
            @param ctime as int
            @return bookmark id as int
        """
        # Find an uniq guid
        while guid is None:
            guid = get_random_string(12)
            if self.exists_guid(guid):
                guid = None

        with SqlCursor(self, True) as sql:
            result = sql.execute("INSERT INTO bookmarks\
                                  (title, uri, popularity, guid, atime, mtime)\
                                  VALUES (?, ?, ?, ?, ?, ?)",
                                 (title, uri.rstrip('/'), 0, guid, atime, 0))
            bookmarks_id = result.lastrowid
            for tag in tags:
                if not tag:
                    continue
                tag_id = self.get_tag_id(tag)
                if tag_id is None:
                    tag_id = self.add_tag(tag)
                sql.execute("INSERT INTO bookmarks_tags\
                             (bookmark_id, tag_id) VALUES (?, ?)",
                            (bookmarks_id, tag_id))
            return bookmarks_id

    def remove(self, bookmark_id):
        """
            Remove bookmark from db
            @param bookmark id as int
        """
        with SqlCursor(self, True) as sql:
            sql.execute("DELETE FROM bookmarks\
                         WHERE rowid=?", (bookmark_id,))
            sql.execute("DELETE FROM bookmarks_tags\
                         WHERE bookmark_id=?", (bookmark_id,))
            sql.execute("DELETE FROM parents\
                         WHERE bookmark_id=?", (bookmark_id,))

    def add_tag(self, tag):
        """
            Add tag to db, return existing if exists
            @param tag as str
            @return tag id as int
        """
        with SqlCursor(self, True) as sql:
            result = sql.execute("INSERT INTO tags\
                                  (title) VALUES (?)",
                                 (tag,))
            return result.lastrowid

    def del_tag(self, tag):
        """
            Add tag to db, return existing if exists
            @param tag as str
        """
        with SqlCursor(self, True) as sql:
            tag_id = self.get_tag_id(tag)
            if tag_id is None:
                return
            sql.execute("DELETE FROM tags\
                         WHERE rowid=?", (tag_id,))
            sql.execute("DELETE FROM bookmarks_tags\
                         WHERE tag_id=?", (tag_id,))

    def rename_tag(self, old, new):
        """
            Rename tag
            @param old as str
            @param new as str
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE tags set title=? WHERE title=?", (new, old))

    def get_tags(self, bookmark_id):
        """
            Get tags for bookmark id
            @param bookmark id as int
            @return [str]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT tags.title\
                                  FROM tags, bookmarks_tags\
                                  WHERE bookmarks_tags.bookmark_id=?\
                                  AND bookmarks_tags.tag_id=tags.rowid\
                                  ORDER BY title COLLATE LOCALIZED",
                                 (bookmark_id,))
            return list(itertools.chain(*result))

    def has_tag(self, bookmark_id, tag):
        """
            Return True if bookmark id as tag
            @param bookmark id as int
            @param tag as str
            @return bool
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT tags.rowid\
                                  FROM tags, bookmarks_tags\
                                  WHERE tags.title=? COLLATE NOCASE\
                                  AND bookmarks_tags.bookmark_id=?\
                                  AND bookmarks_tags.tag_id=tags.rowid",
                                 (tag, bookmark_id))
            v = result.fetchone()
            if v is not None:
                return True
            return False

    def get_id(self, uri):
        """
            Get id for uri
            @param uri as str
            @return id as int
        """
        if uri is None:
            return None
        with SqlCursor(self) as sql:
            parsed = urlparse(uri.rstrip('/'))
            if not parsed.netloc:
                return None
            search = "%{}{}%".format(parsed.netloc, parsed.path)
            result = sql.execute("SELECT rowid\
                                  FROM bookmarks\
                                  WHERE uri like ?", (search,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def get_id_by_guid(self, guid):
        """
            Get id for guid
            @param guid as str
            @return id as int
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT rowid\
                                  FROM bookmarks\
                                  WHERE guid=?", (guid,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def get_ids_for_mtime(self, mtime):
        """
            Get ids that need to be synced related to mtime
            @param mtime as int
            @return [int]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT rowid\
                                  FROM bookmarks\
                                  WHERE mtime > ?\
                                  AND uri != guid", (mtime,))
            return list(itertools.chain(*result))

    def get_parent_guid(self, bookmark_id):
        """
            Get parent for bookmark
            @param bookmark id as int
            @return guid as str
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT parent_guid\
                                  FROM parents\
                                  WHERE bookmark_id=?", (bookmark_id,))
            v = result.fetchone()
            if v is not None and v[0] is not None:
                return v[0]
            return "unfiled"

    def get_parent_name(self, bookmark_id):
        """
            Get parent for bookmark
            @param bookmark id as int
            @return name as str
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT parent_name\
                                  FROM parents\
                                  WHERE bookmark_id=?", (bookmark_id,))
            v = result.fetchone()
            if v is not None and v[0] is not None:
                return v[0]
            return ""

    def get_title(self, bookmark_id):
        """
            Get bookmark title
            @param bookmark id as int
            @return title as str
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT title\
                                  FROM bookmarks\
                                  WHERE rowid=?", (bookmark_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return ""

    def get_uri(self, bookmark_id):
        """
            Get bookmark uri
            @param bookmark id as int
            @return uri as str
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT uri\
                                  FROM bookmarks\
                                  WHERE rowid=?", (bookmark_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return ""

    def get_guid(self, bookmark_id):
        """
            Get bookmark guid
            @param bookmark id as int
            @return guid as str
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT guid\
                                  FROM bookmarks\
                                  WHERE rowid=?", (bookmark_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def get_guids(self):
        """
            Get all guids
            @return guids as [str]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT guid FROM bookmarks")
            return list(itertools.chain(*result))

    def get_children(self, guid):
        """
            Get guid children
            @param guid as str
            @return [str]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT bookmarks.guid\
                                  FROM bookmarks, parents\
                                  WHERE parents.parent_guid=?\
                                  AND parents.bookmark_id=bookmarks.rowid\
                                  ORDER BY position ASC", (guid,))
            return list(itertools.chain(*result))

    def get_mtime(self, bookmark_id):
        """
            Get bookmark mtime
            @param bookmark id as int
            @return mtime as int
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT mtime\
                                  FROM bookmarks\
                                  WHERE rowid=?", (bookmark_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return 0

    def get_position(self, bookmark_id):
        """
            Get bookmark position
            @param bookmark id as int
            @return position as int
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT position\
                                  FROM bookmarks\
                                  WHERE rowid=?", (bookmark_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return 0

    def get_tag_id(self, title):
        """
            Get tag id
            @param title as str
            @return tag id as int
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT rowid\
                                  FROM tags\
                                  WHERE title=? COLLATE NOCASE", (title,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def get_tag_title(self, tag_id):
        """
            Get tag id title
            @param tag id as int
            @return title as str
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT tags.title\
                                  FROM tags\
                                  WHERE id=?", (tag_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def get_all_tags(self):
        """
            Get all tags
            @return [rowid, str]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT rowid, title\
                                  FROM tags\
                                  ORDER BY title COLLATE LOCALIZED")
            return list(result)

    def get_bookmarks(self, tag_id=None):
        """
            Get all bookmarks
            @param tag id as int
            @return [(id, title, uri)]
        """
        with SqlCursor(self) as sql:
            if tag_id is None:
                result = sql.execute("\
                                SELECT bookmarks.rowid,\
                                       bookmarks.uri,\
                                       bookmarks.title\
                                FROM bookmarks\
                                ORDER BY bookmarks.popularity DESC")
            else:
                result = sql.execute("\
                                SELECT bookmarks.rowid,\
                                       bookmarks.uri,\
                                       bookmarks.title\
                                FROM bookmarks, bookmarks_tags\
                                WHERE bookmarks.rowid=\
                                      bookmarks_tags.bookmark_id\
                                      AND bookmarks_tags.tag_id=?\
                                      AND bookmarks.guid != bookmarks.uri\
                                ORDER BY bookmarks.popularity DESC", (tag_id,))
            return list(result)

    def get_populars(self, limit):
        """
            Get popular bookmarks
            @param limit as bool
            @return [(id, title, uri)]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("\
                            SELECT bookmarks.rowid,\
                                   bookmarks.uri,\
                                   bookmarks.title\
                            FROM bookmarks\
                            WHERE popularity!=0\
                            AND bookmarks.guid != bookmarks.uri\
                            ORDER BY bookmarks.popularity DESC,\
                            bookmarks.atime DESC\
                            LIMIT ?", (limit,))
            return list(result)

    def get_unclassified(self):
        """
            Get bookmarks without tag
            @return [(id, title, uri)]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("\
                            SELECT bookmarks.rowid,\
                                   bookmarks.uri,\
                                   bookmarks.title\
                            FROM bookmarks\
                            WHERE NOT EXISTS (\
                                SELECT bookmark_id FROM bookmarks_tags\
                                WHERE bookmark_id=bookmarks.rowid)\
                            AND bookmarks.guid != bookmarks.uri\
                            ORDER BY bookmarks.popularity DESC")
            return list(result)

    def get_recents(self):
        """
            Get recents bookmarks
            @return [(id, title, uri)]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT bookmarks.rowid,\
                                  bookmarks.uri,\
                                  bookmarks.title\
                                  FROM bookmarks\
                                  WHERE bookmarks.guid != bookmarks.uri\
                                  ORDER BY bookmarks.mtime DESC")
            return list(result)

    def get_popularity(self, bookmark_id):
        """
            Get popularity for bookmark id
            @param bookmark_id as int
            @return popularity as int
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT popularity\
                                  FROM bookmarks\
                                  WHERE rowid=?", (bookmark_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return 0

    def get_higher_popularity(self):
        """
            Get higher available popularity
            @return int
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT popularity\
                                  FROM bookmarks\
                                  ORDER BY POPULARITY DESC LIMIT 1")
            v = result.fetchone()
            if v is not None:
                return v[0]
            return 0

    def get_avg_popularity(self):
        """
            Return avarage popularity
            @return avarage popularity as int
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT AVG(popularity)\
                                  FROM (SELECT popularity\
                                        FROM bookmarks\
                                        ORDER BY POPULARITY DESC LIMIT 100)")
            v = result.fetchone()
            if v and v[0] > 5:
                return v[0]
            return 5

    def get_startup(self, bookmark_id):
        """
            True if bookmark needs to be loaded at startup
            @param bookmark_id as int
            @return bool
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT startup\
                                  FROM bookmarks\
                                  WHERE rowid=?", (bookmark_id,))
            v = result.fetchone()
            if v is not None:
                return bool(v[0])
            return False

    def get_startup_uris(self):
        """
            Get URIS to load at startup
            @param bookmark_id as int
            @return [str]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT uri FROM bookmarks\
                                  WHERE startup=1")
            return list(itertools.chain(*result))

    def set_guid(self, bookmark_id, guid):
        """
            Set bookmark guid
            @param bookmark_id as int
            @param guid as str
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE bookmarks\
                         SET guid=?\
                         WHERE rowid=?", (bookmark_id, guid))

    def set_title(self, bookmark_id, title):
        """
            Set bookmark title
            @param bookmark id as int
            @param title as str
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE bookmarks\
                         SET title=?\
                         WHERE rowid=?", (title, bookmark_id,))

    def set_uri(self, bookmark_id, uri):
        """
            Set bookmark uri
            @param bookmark id as int
            @param uri as str
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE bookmarks\
                         SET uri=?\
                         WHERE rowid=?", (uri.rstrip('/'), bookmark_id,))

    def set_popularity(self, bookmark_id, popularity):
        """
            Set bookmark popularity
            @param bookmark id as int
            @param popularity as int
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE bookmarks\
                         SET popularity=?\
                         WHERE rowid=?", (popularity, bookmark_id,))

    def set_startup(self, bookmark_id, startup):
        """
            Set bookmark startup value
            @param bookmark id as int
            @param startup as bool
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE bookmarks\
                         SET startup=?\
                         WHERE rowid=?", (startup, bookmark_id,))

    def set_parent(self, bookmark_id, parent_guid, parent_name):
        """
            Set parent id for bookmark
            @param bookmark_id as int
            @param parent_guid as str
            @param parent_name as str
        """
        with SqlCursor(self, True) as sql:
            result = sql.execute("SELECT parent_guid\
                                  FROM parents\
                                  WHERE bookmark_id=?", (bookmark_id,))
            v = result.fetchone()
            if v is None or v[0] is None:
                sql.execute("INSERT INTO parents\
                             (bookmark_id, parent_guid, parent_name)\
                             VALUES (?, ?, ?)",
                            (bookmark_id, parent_guid, parent_name))
            else:
                sql.execute("UPDATE parents\
                             SET parent_guid=?, parent_name=?\
                             WHERE bookmark_id=?",
                            (parent_guid, parent_name, bookmark_id))

    def set_access_time(self, uri, atime):
        """
            Set bookmark access time
            @param uri as str
            @param atime as int
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE bookmarks\
                         SET atime=? where uri=?", (atime, uri.rstrip('/')))

    def set_mtime(self, bookmark_id, mtime):
        """
            Set bookmark sync time
            @param bookmark id as int
            @param mtime as int
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE bookmarks\
                         SET mtime=? where rowid=?", (mtime, bookmark_id))

    def set_position(self, bookmark_id, position):
        """
            Set bookmark position
            @param bookmark id as int
            @param mtime as int
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE bookmarks\
                         SET position=? where rowid=?", (position,
                                                         bookmark_id))

    def get_match(self, uri, ssl_force=False):
        """
            Try to get best uri matching
            @parma uri as str
            @param ssl_force as bool
            @return str
        """
        with SqlCursor(self) as sql:
            if ssl_force:
                filter = ("https://%{}%".format(uri),)
            else:
                filter = ("http%://%{}%".format(uri),)
            result = sql.execute("SELECT uri\
                                  FROM bookmarks\
                                  WHERE uri like ?\
                                  ORDER BY length(uri) ASC\
                                  LIMIT 1", filter)
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def set_tag_title(self, tag_id, title):
        """
            Set tag id title
            @param tag id as int
            @parma title as str
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE tags SET title=? WHERE id=?", (title, tag_id,))

    def set_more_popular(self, uri):
        """
            Increment bookmark popularity
            @param uri as str
        """
        with SqlCursor(self, True) as sql:
            uri = uri.rstrip('/')
            result = sql.execute("SELECT popularity FROM bookmarks\
                                  WHERE uri=?", (uri,))
            v = result.fetchone()
            if v is not None:
                sql.execute("UPDATE bookmarks set popularity=?\
                             WHERE uri=?", (v[0] + 1, uri))

    def add_tag_to(self, tag_id, bookmark_id):
        """
            Add tag to bookmark
            @param tag id as int
            @param bookmark id as int
        """
        with SqlCursor(self, True) as sql:
            sql.execute("INSERT INTO bookmarks_tags\
                         (bookmark_id, tag_id) VALUES (?, ?)",
                        (bookmark_id, tag_id))

    def del_tag_from(self, tag_id, bookmark_id):
        """
            Remove tag from bookmark
            @param tag id as int
            @param bookmark id as int
        """
        with SqlCursor(self, True) as sql:
            sql.execute("DELETE from bookmarks_tags\
                         WHERE bookmark_id=? and tag_id=?",
                        (bookmark_id, tag_id))

    def clean_tags(self):
        """
            Remove orphan tags
        """
        with SqlCursor(self, True) as sql:
            sql.execute("DELETE from tags\
                         WHERE NOT EXISTS (\
                            SELECT bookmarks_tags.rowid\
                            FROM bookmarks, bookmarks_tags\
                            WHERE tags.rowid = bookmarks_tags.tag_id\
                            AND bookmarks.rowid = bookmarks_tags.bookmark_id)")

    def reset_popularity(self, uri):
        """
            Reset popularity for uri
            @param uri as str
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE bookmarks SET popularity=0 WHERE uri=?",
                        (uri,))

    def is_empty(self):
        """
            True if db is empty
            @return bool
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT rowid FROM bookmarks LIMIT 1")
            v = result.fetchone()
            if v is None:
                return True
            return False

    def import_html(self, path):
        """
            Import html bookmarks
            @param path as str
        """
        try:
            from bs4 import BeautifulSoup
            SqlCursor.add(self)
            f = Gio.File.new_for_path(path)
            if not f.query_exists():
                return
            (status, content, tag) = f.load_contents(None)
            if status:
                data = content.decode("utf-8")
                soup = BeautifulSoup(data, "html.parser")
                parent_name = ""
                position = 0
                for dt in soup.findAll("dt"):
                    h3 = dt.find("h3")
                    if h3 is not None:
                        parent_name = h3.contents[0]
                        continue
                    else:
                        a = dt.find("a")
                        uri = a.get("href")
                        if a.get("tags") is None:
                            tags = [parent_name]
                        else:
                            tags = [a.get("tags")]
                        title = a.contents[0]
                        if uri is None:
                            parent_name = title
                            continue
                        elif not uri.startswith('http') or not title:
                            continue
                        uri = uri.rstrip('/')
                        rowid = self.get_id(uri)
                        if rowid is None:
                            if not tags:
                                tags = [parent_name]
                            # Add bookmark
                            bookmark_id = self.add(title, uri, None,
                                                   tags, 0)
                            # Set position
                            self.set_position(bookmark_id, position)
                            position += 1
            SqlCursor.remove(self)
        except Exception as e:
            Logger.error("DatabaseBookmarks::import_html(): %s", e)

    def import_chromium(self, chrome):
        """
            Chromium/Chrome importer
            As Eolie doesn't sync with Chromium, we do not handle parent
            guid and just import parents as tags
            @param chrome as bool
        """
        try:
            SqlCursor.add(self)
            import json
            homedir = GLib.get_home_dir()
            if chrome:
                path = homedir + "/.config/chrome/Default/Bookmarks"
            else:
                path = homedir + "/.config/chromium/Default/Bookmarks"
            status = False
            f = Gio.File.new_for_path(path)
            if f.query_exists():
                (status, content, tag) = f.load_contents(None)
            if status:
                data = content.decode("utf-8")
                j = json.loads(data)
                parents = []
                # Setup initial parents
                for root in j["roots"]:
                    parents.append(("", j["roots"][root]["children"]))
                # Walk parents and children
                while parents:
                    (parent_name, children) = parents.pop(0)
                    bookmarks = []
                    for child in children:
                        if child["type"] == "folder":
                            parents.append((child["name"], child["children"]))
                        elif child["type"] == "url":
                            bookmarks.append((child["name"],
                                              child["url"]))
                    position = 0
                    for bookmark in bookmarks:
                        tags = [parent_name]
                        title = bookmark[0]
                        uri = bookmark[1]
                        if not uri.startswith('http') or not title:
                            continue
                        uri = uri.rstrip('/')
                        rowid = self.get_id(uri)
                        if rowid is None:
                            # Add bookmark
                            bookmark_id = self.add(title, uri, None,
                                                   tags, 0)
                            # Set position
                            self.set_position(bookmark_id, position)
                            position += 1
            SqlCursor.remove(self)
        except Exception as e:
            Logger.error("DatabaseBookmarks::import_chromium(): %s", e)

    def import_firefox(self, profile):
        """
            Mozilla Firefox importer
            @param profile as str
        """
        try:
            SqlCursor.add(self)
            path = "%s/.mozilla/firefox/%s/places.sqlite" % \
                (GLib.get_home_dir(),
                 profile)
            f = Gio.File.new_for_path(path)
            if f.query_exists():
                c = sqlite3.connect(path, 600.0)
                # Add bookmarks
                bookmarks = self.__get_firefox_bookmarks(c)
                for (title, uri, parent_name, bookmark_guid,
                     parent_guid, position) in bookmarks:
                    tags = self.__get_tags_for_firefox_bookmark(c,
                                                                bookmark_guid)
                    bookmark_guid = self.__clean_guid(bookmark_guid)
                    parent_guid = self.__clean_guid(parent_guid)
                    if not uri.startswith('http') or not title:
                        continue
                    uri = uri.rstrip('/')
                    rowid = self.get_id(uri)
                    if rowid is None:
                        # If bookmark is not tagged, we use parent name
                        if not tags:
                            tags = [parent_name]
                        # Bookmarks and folder
                        bookmark_id = self.add(title, uri, bookmark_guid,
                                               tags, 0)
                        self.set_parent(bookmark_id, parent_guid,
                                        parent_name)
                        self.set_position(bookmark_id, position)
                # Add folders, we need to get them
                # as Firefox needs children order
                parents = self.__get_firefox_parents(c)
                for (title, parent_name, bookmark_guid,
                     parent_guid, position) in parents:
                    bookmark_guid = self.__clean_guid(bookmark_guid)
                    parent_guid = self.__clean_guid(parent_guid)
                    if not title or bookmark_guid == "root":
                        continue
                    uri = bookmark_guid
                    rowid = self.get_id(uri)
                    if rowid is None:
                        # Bookmarks and folder
                        bookmark_id = self.add(title, uri, bookmark_guid,
                                               [], 0)
                        self.set_parent(bookmark_id, parent_guid,
                                        parent_name)
                        self.set_position(bookmark_id, position)
            SqlCursor.remove(self)
        except Exception as e:
            Logger.error("DatabaseBookmarks::import_firefox(): %s", e)

    def exists_guid(self, guid):
        """
            Check if guid exists in db
            @return bool
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT guid FROM bookmarks\
                                  WHERE guid=?", (guid,))
            v = result.fetchone()
            return v is not None

    def search(self, search, limit):
        """
            Search string in db (uri and title)
            @param search as str
            @param limit as int
            @return [(id, title, uri)] as [(int, str, str)]
        """
        words = search.lower().split()
        items = []
        with SqlCursor(self) as sql:
            filters = ()
            for word in words:
                filters += ("%" + word + "%", "%" + word + "%")
            filters += (limit,)

            # Search items matching all words
            request = "SELECT rowid, title, uri\
                       FROM bookmarks WHERE "
            words_copy = list(words)
            while words_copy:
                word = words_copy.pop(0)
                if word:
                    request += " (title LIKE ? OR uri LIKE ?) AND"
            request += " guid != uri ORDER BY length(uri) ASC LIMIT ?"

            result = sql.execute(request, filters)
            items = list(result)
        return items

    def get_cursor(self):
        """
            Return a new sqlite cursor
        """
        try:
            c = sqlite3.connect(self.DB_PATH, 600.0)
            c.create_collation('LOCALIZED', LocalizedCollation())
            c.create_function("noaccents", 1, noaccents)
            return c
        except Exception as e:
            Logger.error("DatabaseBookmarks::get_cursor(): %s", e)
            exit(-1)

#######################
# PRIVATE             #
#######################
    def __get_firefox_bookmarks(self, c):
        """
            Return firefox bookmarks
            @param c as Sqlite cursor
            @return (title, url, parent title, guid, parent guid, position)
             as (str, str, str, str, str, int)
        """
        result = c.execute("SELECT bookmarks.title,\
                                   moz_places.url,\
                                   parent.title,\
                                   bookmarks.guid,\
                                   parent.guid,\
                                   bookmarks.position\
                            FROM moz_bookmarks AS bookmarks,\
                                 moz_bookmarks AS parent,\
                                 moz_places\
                            WHERE bookmarks.fk=moz_places.id\
                            AND parent.id=bookmarks.parent\
                            AND bookmarks.type=1")
        return list(result)

    def __get_tags_for_firefox_bookmark(self, c, guid):
        """
            Return firefox bookmarks
            @param c as Sqlite cursor
            @param guid as str
            @return (title, url, parent title, guid, parent guid, position)
             as (str, str, str, str, str, int)
        """
        result = c.execute("SELECT parent.title\
                            FROM moz_bookmarks AS bookmarks,\
                                 moz_bookmarks AS tag,\
                                 moz_bookmarks AS parent\
                            WHERE bookmarks.fk=tag.fk\
                            AND tag.fk=bookmarks.fk\
                            AND tag.title is null\
                            AND parent.id=tag.parent\
                            AND bookmarks.guid=?", (guid,))
        return list(itertools.chain(*result))

    def __get_firefox_parents(self, c):
        """
            Return firefox parents
            @param c as Sqlite cursor
            @return (title, parent title, guid, parent guid, position)
             as (str, str, str, str, str, int)
        """
        result = c.execute("SELECT bookmarks.title,\
                                   parent.title,\
                                   bookmarks.guid,\
                                   parent.guid,\
                                   bookmarks.position\
                            FROM moz_bookmarks AS bookmarks,\
                                 moz_bookmarks AS parent\
                            WHERE parent.id=bookmarks.parent\
                            AND bookmarks.type=2")
        return list(result)

    def __clean_guid(self, guid):
        """
            Clean guid to match sync API
            @param guid as str
            @return str
        """
        if guid == "root________":
            return "places"
        elif guid == "menu________":
            return "menu"
        elif guid == "toolbar_____":
            return "toolbar"
        elif guid == "unfiled_____":
            return "unfiled"
        elif guid == "mobile______":
            return "mobile"
        else:
            return guid



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, GObject, Pango

from gettext import gettext as _
from urllib.parse import urlparse

from eolie.search import Search
from eolie.define import App, MARGIN_SMALL
from eolie.utils import emit_signal
from eolie.logger import Logger


class Item(GObject.GObject):
    name = GObject.Property(type=str,
                            default="")
    uri = GObject.Property(type=str,
                           default="")
    search = GObject.Property(type=str,
                              default="")
    keyword = GObject.Property(type=str,
                               default="")
    encoding = GObject.Property(type=str,
                                default="")
    bang = GObject.Property(type=str,
                            default="")

    def __init__(self):
        GObject.GObject.__init__(self)


class Row(Gtk.ListBoxRow):
    """
        A cookie row
    """

    def __init__(self, item):
        """
            Init row
            @param item as Item
        """
        Gtk.ListBoxRow.__init__(self)
        self.__item = item
        self.__label = Gtk.Label.new(item.get_property("name"))
        self.__label.set_ellipsize(Pango.EllipsizeMode.END)
        self.__label.set_property("margin", MARGIN_SMALL)
        self.__label.set_property("halign", Gtk.Align.START)
        self.__label.show()
        self.add(self.__label)

    def set_name(self, name):
        """
            Set name
            @param name as str
        """
        self.__label.set_text(name)

    @property
    def is_valid(self):
        """
            True if item is valid
            @return bool
        """
        uri = self.__item.get_property("uri")
        search = self.__item.get_property("search")
        parsed = urlparse(uri)
        if parsed.scheme not in ["http", "https"] or not parsed.netloc:
            return False
        parsed = urlparse(search)
        if parsed.scheme not in ["http", "https"] or search.find("%s") == -1:
            return False
        return True

    @property
    def item(self):
        """
            Get associated item
            @return item
        """
        return self.__item


class SearchEngineDialog(Gtk.Bin):
    """
        A search engine dialog
        THANKS TO EPIPHANY DEVS FOR UI FILE!
    """
    __gsignals__ = {
        "destroy-me": (GObject.SignalFlags.RUN_FIRST, None, ())
    }

    def __init__(self):
        """
            Init widget
        """
        Gtk.Bin.__init__(self)
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/DialogSearchEngine.ui")
        self.__edit_box = builder.get_object("edit_box")
        self.__listbox = builder.get_object("list_box")
        self.__name_entry = builder.get_object("name_entry")
        self.__uri_entry = builder.get_object("uri_entry")
        self.__search_entry = builder.get_object("search_entry")
        self.__bang_entry = builder.get_object("bang_entry")
        self.__default_switch = builder.get_object("default_switch")
        self.__add_button = builder.get_object("add_button")
        self.__remove_button = builder.get_object("remove_button")
        self.__remove_button.set_sensitive(False)
        builder.connect_signals(self)
        self.__populate()
        self.connect("destroy", self.__on_destroy)
        self.add(builder.get_object("widget"))

#######################
# PROTECTED           #
#######################
    def _on_back_clicked(self, button):
        """
            Ask to be destroyed
            @param button as Gtk.Button
        """
        emit_signal(self, "destroy-me")

    def _on_default_switch_state_set(self, switch, state):
        """
            Update engine state
            @param switch as Gtk.Switch
            @param state as bool
        """
        if state:
            row = self.__listbox.get_selected_row()
            if row is not None:
                name = row.item.get_property("name")
                App().settings.set_value("search-engine",
                                         GLib.Variant("s", name))

    def _on_add_button_clicked(self, button):
        """
            Add a new engine
            @param button as Gtk.Button
        """
        # Only one New engine
        for child in self.__listbox.get_children():
            if child.item.name == _("New engine"):
                return

        item = Item()
        item.set_property("name", _("New engine"))
        item.set_property("uri", "")
        item.set_property("search", "")
        item.set_property("bang", "")
        child = Row(item)
        child.show()
        self.__listbox.add(child)
        self.__listbox.select_row(child)
        self.__remove_button.set_sensitive(True)

    def _on_remove_button_clicked(self, button):
        """
            Remove engine
            @param button as Gtk.Button
        """
        row = self.__listbox.get_selected_row()
        if row is not None:
            row.destroy()

    def _on_row_selected(self, listbox, row):
        """
            Update entries
            @param listbox as Gtk.ListBox
            @param row as Row
        """
        if row is None:
            children = listbox.get_children()
            if children:
                GLib.idle_add(listbox.select_row, children[0])
            else:
                self.__remove_button.set_sensitive(False)
                self.__edit_box.set_sensitive(False)
            return
        self.__remove_button.set_sensitive(True)
        self.__edit_box.set_sensitive(True)
        self.__name_entry.set_text(row.item.get_property("name"))
        self.__uri_entry.set_text(row.item.get_property("uri"))
        self.__search_entry.set_text(row.item.get_property("search"))
        self.__bang_entry.set_text(row.item.get_property("bang"))
        default_search_engine = App().settings.get_value(
            "search-engine").get_string()
        self.__default_switch.set_active(default_search_engine ==
                                         row.item.get_property("name"))

    def _on_name_entry_changed(self, entry):
        """
            Update search engine name
            @param entry as Gtk.Entry
        """
        new_name = entry.get_text()
        row = self.__listbox.get_selected_row()
        if row is None or not new_name:
            return
        # Update search engine if needed
        name = row.item.get_property("name")
        if name == App().settings.get_value("search-engine").get_string():
            App().settings.set_value("search-engine",
                                     GLib.Variant("s", new_name))
        row.item.set_property("name", new_name)
        row.set_name(new_name)

    def _on_uri_entry_changed(self, entry):
        """
            Update search engine URI
            @param entry as Gtk.Entry
        """
        uri = entry.get_text()
        row = self.__listbox.get_selected_row()
        if row is None or not uri:
            return
        row.item.set_property("uri", uri)
        parsed = urlparse(uri)
        if parsed.scheme not in ["http", "https"] or not parsed.netloc:
            entry.get_style_context().add_class("invalid-entry")
        else:
            entry.get_style_context().remove_class("invalid-entry")

    def _on_search_entry_changed(self, entry):
        """
            Update search engine search URI
            @param entry as Gtk.Entry
        """
        search = entry.get_text()
        row = self.__listbox.get_selected_row()
        if row is None or not search:
            return
        row.item.set_property("search", search)
        parsed = urlparse(search)
        if parsed.scheme not in ["http", "https"] or search.find("%s") == -1:
            entry.get_style_context().add_class("invalid-entry")
        else:
            entry.get_style_context().remove_class("invalid-entry")

    def _on_bang_entry_changed(self, entry):
        """
            Update search engine bang
            @param entry as Gtk.Entry
        """
        row = self.__listbox.get_selected_row()
        if row is None or not entry.get_text():
            return
        row.item.set_property("bang", entry.get_text())

    def _on_bang_key_press_event(self, entry, event):
        """
            Validate bang value
            @param entry as Gtk.Entry
            @param event as Gdk.EventKey
        """
        row = self.__listbox.get_selected_row()
        if row is None:
            return
        # Get all bang
        bangs = []
        for child in self.__listbox.get_children():
            if child == row:
                continue
            bang = child.item.get_property("bang")
            if bang:
                bangs.append(bang)
        if event.string in bangs:
            return True

#######################
# PRIVATE             #
#######################
    def __populate(self):
        """
            Populate the view with engines
        """
        search = Search()
        engines = search.engines
        try:
            # First load static engines
            for key in engines.keys():
                item = Item()
                item.set_property("name", key)
                item.set_property("uri", engines[key][0])
                item.set_property("search", engines[key][1])
                item.set_property("keyword", engines[key][2])
                item.set_property("encoding", engines[key][3])
                item.set_property("bang", engines[key][4])
                child = Row(item)
                child.show()
                self.__listbox.add(child)
        except Exception as e:
            Logger.error("DialogSearchEngine::__populate(): %s", e)

    def __on_destroy(self, widget):
        """
            Save engines
            @param widget as Gtk.Widget
        """
        # Save engines
        engines = {}
        for child in self.__listbox.get_children():
            if child.is_valid:
                name = child.item.get_property("name")
                uri = child.item.get_property("uri")
                search = child.item.get_property("search")
                keyword = child.item.get_property("keyword")
                encoding = child.item.get_property("encoding")
                bang = child.item.get_property("bang")
                if name and search:
                    engines[name] = [uri, search, keyword, encoding, bang]
        App().search.save_engines(engines)
        App().search.update_default_engine()



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio


class Settings(Gio.Settings):
    """
        Eolie settings
    """

    def __init__(self):
        """
            Init settings
        """
        Gio.Settings.__init__(self)

    def new():
        """
            Return a new Settings object
        """
        settings = Gio.Settings.new("org.gnome.Eolie")
        settings.__class__ = Settings
        return settings



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('WebKit2', '4.1')
gi.require_version('Soup', '3.0')
gi.require_version('Secret', '1')
gi.require_version('GtkSpell', '3.0')
gi.require_version("Handy", "1")
from gi.repository import Gtk, Gio, GLib, Gdk, WebKit2, Handy

from threading import current_thread
from gettext import gettext as _
from pickle import dump, load
from urllib.parse import urlparse
from time import time
from getpass import getuser
from signal import signal, SIGINT, SIGTERM
import gc
import os
import shutil

from eolie.application_night import NightApplication
from eolie.settings import Settings
from eolie.window import Window
from eolie.art import Art
from eolie.content_blocker_ad import AdContentBlocker
from eolie.content_blocker_popups import PopupsContentBlocker
from eolie.content_blocker_images import ImagesContentBlocker
from eolie.content_blocker_medias import MediasContentBlocker
from eolie.content_blocker_scripts import ScriptsContentBlocker
from eolie.content_blocker_phishing import PhishingContentBlocker
from eolie.database_history import DatabaseHistory
from eolie.database_bookmarks import DatabaseBookmarks
from eolie.database_settings import DatabaseSettings
from eolie.sqlcursor import SqlCursor
from eolie.search import Search
from eolie.download_manager import DownloadManager
from eolie.menu_pages import PagesMenu
from eolie.helper_task import TaskHelper
from eolie.define import EOLIE_DATA_PATH, TimeSpan, TimeSpanValues, LoadingType
from eolie.define import StartPage
from eolie.utils import is_unity, wanted_loading_type
from eolie.logger import Logger
from eolie.webview_state import WebViewState


class Application(Gtk.Application, NightApplication):
    """
        Eolie application
    """

    __FAVICONS_PATH = "/tmp/eolie_%s" % getuser()

    def __init__(self, version, data_dir, app_id):
        """
            Create application
            @param version as str
            @param data_dir as str
            @param app_id as str
        """
        self.__version = version
        self.__state_cache = []
        self.__data_dir = data_dir
        self.__app_id = app_id
        self.__content_blockers = []
        signal(SIGINT, lambda a, b: self.quit())
        signal(SIGTERM, lambda a, b: self.quit())
        # Set main thread name
        # We force it to current python 3.6 name, to be sure in case of
        # change in python
        current_thread().setName("MainThread")
        # First check WebKit2 version
        if WebKit2.MINOR_VERSION < 20:
            exit("You need WebKit2GTK >= 2.20")
        Gtk.Application.__init__(
            self,
            application_id=app_id,
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE | Gio.ApplicationFlags.HANDLES_OPEN)
        self.set_property("register-session", True)
        # Fix proxy for python
        proxy = GLib.environ_getenv(GLib.get_environ(), "all_proxy")
        if proxy is not None and proxy.startswith("socks://"):
            proxy = proxy.replace("socks://", "socks4://")
            from os import environ
            environ["all_proxy"] = proxy
            environ["ALL_PROXY"] = proxy
        # Flatpak Hacks
        if GLib.file_test("/app", GLib.FileTest.EXISTS):
            # Set /tmp for GLib, /tmp not accessible in flatpak
            tmp = GLib.environ_getenv(GLib.get_environ(), "XDG_RUNTIME_DIR")
            GLib.setenv("TMPDIR", "%s/app/org.gnome.Eolie" % tmp, True)
        self.sync_worker = None  # Not initialised
        self.show_tls = False
        self.cursors = {}
        GLib.set_application_name('Eolie')
        GLib.set_prgname('org.gnome.Eolie')
        self.add_main_option("private", b'p', GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE, "Add a private page",
                             None)
        self.add_main_option("new", b'n', GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE, "Add a new window",
                             None)
        self.add_main_option("disable-artwork-cache", b'a',
                             GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE, "Do not use cache for art",
                             None)
        self.add_main_option("show-tls", b't', GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE, "Show TLS info",
                             None)
        self.connect("activate", self.__on_activate)
        self.connect("open", self.__on_open)
        self.connect("handle-local-options", self.__on_handle_local_options)
        self.connect("command-line", self.__on_command_line)

        new_window = Gio.SimpleAction.new("new-window", None)
        new_window.connect("activate", self.__on_new_window_action)
        self.add_action(new_window)

        private_page = Gio.SimpleAction.new("private-page", None)
        private_page.connect("activate", self.__on_private_page_action)
        self.add_action(private_page)

        self.register(None)
        if self.get_is_remote():
            Gdk.notify_startup_complete()
        if GLib.environ_getenv(GLib.get_environ(), "DEBUG_LEAK") is not None:
            gc.set_debug(gc.DEBUG_LEAK)
        else:
            gc.disable()

    def get_new_window(self):
        """
            Return a new window
            @return Window
        """
        size = (800, 600)
        is_maximized = False
        windows = self.get_windows()
        if windows:
            active_window = self.active_window
            size = active_window.size
            is_maximized = active_window.is_maximized()
        window = Window(size, is_maximized)
        window.connect('delete-event', self.__on_delete_event)
        window.show()
        return window

    def set_setting(self, key, value):
        """
            Set setting for all view
            @param key as str
            @param value as GLib.Variant
        """
        for window in self.get_windows():
            for webview in window.container.webviews:
                webview.set_setting(key, value)

    def update_unity_badge(self, fraction=None):
        """
            Update unity badge count/fraction
            @param fraction as double
        """
        if self.__unity is not None:
            if fraction is None:
                count = 0
                for window in self.get_windows():
                    count += len(window.container.pages_manager.children)
                self.__unity.set_property("count", count)
                self.__unity.set_property("count_visible", True)
            else:
                self.__unity.set_property("progress", fraction)
                self.__unity.set_property("progress_visible", fraction != 1.0)

    def quit(self, vacuum=False):
        """
            Quit application
            @param vacuum as bool
        """
        self.__save_state()
        for window in self.windows:
            window.hide()
        # Stop pending tasks
        self.download_manager.cancel()
        for content_blocker in self.__content_blockers:
            content_blocker.stop()
        # Clear history
        active_id = str(self.settings.get_enum("history-storage"))
        if active_id != TimeSpan.FOREVER:
            atime = time()
            if active_id != TimeSpan.NEVER:
                atime -= TimeSpanValues[active_id] / 1000000
            self.history.clear_to(int(atime))

        if self.sync_worker is not None:
            self.sync_worker.stop()
            self.sync_worker.save_pendings()

        if os.path.exists(self.favicons_path):
            shutil.rmtree(self.favicons_path)

        if GLib.environ_getenv(GLib.get_environ(), "DEBUG_LEAK") is not None:
            gc.collect()
            for x in gc.garbage:
                s = str(x)
                print(type(x), "\n  ", s)
        if vacuum:
            self.task_helper.run(
                        self.__vacuum,
                        callback=(lambda x: Gio.Application.quit(self),))
        else:
            Gio.Application.quit(self)

    def get_content_blocker(self, name):
        """
            Get content blocker by name
            @param name as str
            @return ContentBlocker
        """
        for content_blocker in self.__content_blockers:
            if content_blocker.name == name:
                return content_blocker

    @property
    def content_filters(self):
        """
            Get content filters
            @return [WebKit2.UserContentFilter]
        """
        filters = []
        for content_blocker in self.__content_blockers:
            if content_blocker.enabled:
                filters.append(content_blocker.filter)
        return filters

    @property
    def start_page(self):
        """
            Get start page
            @return uri as str
        """
        value = self.settings.get_enum("start-page")
        if value in [StartPage.POPULARITY_HISTORY,
                     StartPage.POPULARITY_BOOKMARKS]:
            start_page = "populars://"
        elif value == StartPage.BLANK:
            start_page = "about:blank"
        elif value == StartPage.SEARCH:
            start_page = self.search.uri
        else:
            start_page = self.settings.get_value(
                "start-page-custom").get_string()
        return start_page

    @property
    def active_window(self):
        """
            Get active window
            @return Window
        """
        return self.get_active_window()

    @property
    def windows(self):
        """
            Get windows
            @return [Window]
        """
        return self.get_windows()

    @property
    def data_dir(self):
        """
            Get data dir
            @return str
        """
        return self.__data_dir

    @property
    def favicons_path(self):
        """
            Cookies sqlite DB path
            @return str
        """
        return self.__FAVICONS_PATH

#######################
# PRIVATE             #
#######################
    def __init(self):
        """
            Init main application
        """
        Handy.init()
        self.settings = Settings.new()
        NightApplication.__init__(self)

        # First init sync worker
        from eolie.firefox_sync import SyncWorker
        if False and SyncWorker.check_modules():
            self.sync_worker = SyncWorker()
            self.sync_worker.pull_loop()
        else:
            self.sync_worker = None
        cssProviderFile = Gio.File.new_for_uri(
            'resource:///org/gnome/Eolie/application.css')
        cssProvider = Gtk.CssProvider()
        cssProvider.load_from_file(cssProviderFile)
        screen = Gdk.Screen.get_default()
        styleContext = Gtk.StyleContext()
        styleContext.add_provider_for_screen(screen, cssProvider,
                                             Gtk.STYLE_PROVIDER_PRIORITY_USER)
        # Devel version, load devel theme
        if self.__app_id is None:
            css = ".progressbar progress { background-color: red; }"
            cssProvider = Gtk.CssProvider()
            cssProvider.load_from_data(css.encode("utf-8"))
            screen = Gdk.Screen.get_default()
            styleContext = Gtk.StyleContext()
            styleContext.add_provider_for_screen(
                screen, cssProvider, Gtk.STYLE_PROVIDER_PRIORITY_USER + 1)
        self.history = DatabaseHistory()
        self.bookmarks = DatabaseBookmarks()
        self.websettings = DatabaseSettings()
        for cls in [AdContentBlocker,
                    PopupsContentBlocker,
                    ImagesContentBlocker,
                    MediasContentBlocker,
                    ScriptsContentBlocker,
                    PhishingContentBlocker]:
            content_blocker = cls()
            content_blocker.connect("set-filter",
                                    self.__on_content_blocker_set_filter)
            content_blocker.connect("unset-filter",
                                    self.__on_content_blocker_unset_filter)
            self.__content_blockers.append(content_blocker)
        self.art = Art()

        # Get a default user agent for search
        settings = WebKit2.Settings.new()
        self.search = Search(settings.get_user_agent())

        self.task_helper = TaskHelper()
        self.download_manager = DownloadManager()
        self.pages_menu = PagesMenu()

        # Check MOZ_PLUGIN_PATH
        if self.settings.get_value('enable-plugins') and\
                not GLib.getenv("MOZ_PLUGIN_PATH"):
            Logger.info("You need to set MOZ_PLUGIN_PATH to use plugins")

        # https://wiki.ubuntu.com/Unity/LauncherAPI
        self.__unity = None
        if is_unity():
            try:
                gi.require_version('Unity', '7.0')
                from gi.repository import Unity
                self.__unity = Unity.LauncherEntry.get_for_desktop_id(
                    "org.gnome.Eolie.desktop")
            except:
                pass

        self.set_accels_for_action("win.shortcut::expose", ["<Alt>e"])
        self.set_accels_for_action("win.shortcut::show_sidebar", ["F9"])
        self.set_accels_for_action("win.shortcut::uri",
                                   ["<Control>l", "<Control>b"])
        self.set_accels_for_action("win.shortcut::new_page", ["<Control>t"])
        self.set_accels_for_action("win.shortcut::last_page",
                                   ["<Control><Shift>t"])
        self.set_accels_for_action("app.new-window", ["<Control>n"])
        self.set_accels_for_action("win.shortcut::private",
                                   ["<Control><Shift>p"])
        self.set_accels_for_action("win.shortcut::close_page", ["<Control>w"])
        self.set_accels_for_action("win.shortcut::quit", ["<Control>q"])
        self.set_accels_for_action("win.shortcut::save", ["<Control><Shift>s"])
        self.set_accels_for_action("win.shortcut::filter", ["<Control>i"])
        self.set_accels_for_action("win.shortcut::reload",
                                   ["<Control>r", "F5"])
        self.set_accels_for_action("win.shortcut::home",
                                   ["<Control>Home"])
        self.set_accels_for_action("win.shortcut::find", ["<Control>f"])
        self.set_accels_for_action("win.shortcut::print", ["<Control>p"])
        self.set_accels_for_action("win.shortcut::source",
                                   ["<Control><Shift>c"])
        self.set_accels_for_action("win.shortcut::history", ["<Control>h"])
        self.set_accels_for_action("win.shortcut::search", ["<Control>k"])
        self.set_accels_for_action("win.shortcut::fullscreen", ["F11"])
        self.set_accels_for_action("win.shortcut::settings",
                                   ["<Control>comma"])
        self.set_accels_for_action("win.shortcut::backward",
                                   ["<Alt>Left", "Back"])
        self.set_accels_for_action("win.shortcut::forward",
                                   ["<Alt>Right", "Forward"])
        self.set_accels_for_action("win.shortcut::next",
                                   ["<Control>Tab",
                                    "<Control>Page_Down"])
        self.set_accels_for_action("win.shortcut::previous",
                                   ["<Control><Shift>Tab",
                                    "<Control>Page_Up"])
        self.set_accels_for_action("win.shortcut::next_site",
                                   ["<Control>twosuperior"])
        self.set_accels_for_action("win.shortcut::previous_site",
                                   ["<Control><Shift>twosuperior"])
        self.set_accels_for_action("win.shortcut::zoom_in",
                                   ["<Control>KP_Add",
                                    "<Control>plus",
                                    "<Control>equal"])
        self.set_accels_for_action("win.shortcut::zoom_out",
                                   ["<Control>KP_Subtract", "<Control>minus"])
        self.set_accels_for_action("win.shortcut::zoom_default",
                                   ["<Control>KP_0", "<Control>0"])
        self.set_accels_for_action("win.shortcut::mse_enabled",
                                   ["<Control>m"])

    def __vacuum(self):
        """
            VACUUM DB
            @thread safe
        """
        try:
            with SqlCursor(self.bookmarks) as sql:
                sql.isolation_level = None
                sql.execute("VACUUM")
                sql.isolation_level = ""
            with SqlCursor(self.history) as sql:
                sql.isolation_level = None
                sql.execute("VACUUM")
                sql.isolation_level = ""
        except Exception as e:
            Logger.error("Application::__vacuum(): %s ", e)
        self.art.vacuum()

    def __save_state(self):
        """
            Save windows state
        """
        try:
            window_states = []
            for window in self.windows:
                state = window.state
                window_states.append(state)
            for window_state in self.__state_cache:
                window_states.append(window_state)
            dump(window_states,
                 open(EOLIE_DATA_PATH + "/session_states.bin", "wb"))
        except Exception as e:
            Logger.error("Application::__save_state(): %s", e)

    def __clean_state_cache(self, window_id):
        """
            Remove window ID from cache
            @param window_id as str
        """
        for state in self.__state_cache:
            if state.wid == window_id:
                self.__state_cache.remove(state)
                break

    def __restore_state(self):
        """
            Restore state
            @return active window as Window
        """
        try:
            pinned_netlocs = self.websettings.get_pinned_netlocs()
            startup_uris = self.bookmarks.get_startup_uris()
            # Add saved webviews
            from eolie.window_state import WindowState, WindowStateStruct
            window_states = load(
                open(EOLIE_DATA_PATH + "/session_states.bin", "rb"))
            state = WindowStateStruct()
            for state in window_states:
                # If webview to restore
                if state.webview_states:
                    window = WindowState.new_from_state(state)
                    for webview_state in state.webview_states:
                        parsed = urlparse(webview_state.uri)
                        if parsed.netloc in pinned_netlocs:
                            pinned_netlocs.remove(parsed.netloc)
                        if webview_state.uri in startup_uris:
                            startup_uris.remove(webview_state.uri)
                        webview = WebViewState.new_from_state(webview_state,
                                                              window)
                        webview.show()
                        webview.set_shown(True)
                        loading_type = wanted_loading_type(
                            len(window.container.webviews))
                        window.container.add_webview(webview, loading_type)
                        session = WebKit2.WebViewSessionState(
                            GLib.Bytes.new(webview_state.session))
                        webview.restore_session_state(session)
                    window.connect("delete-event", self.__on_delete_event)
                    window.show()
            # Add a default window
            if not self.windows:
                window = WindowState.new_from_state(state)
                window.connect("delete-event", self.__on_delete_event)
                window.show()
            # Add pinned websites
            for netloc in pinned_netlocs:
                webview = window.container.add_webview_for_uri(
                                                     "http://%s" % netloc,
                                                     LoadingType.BACKGROUND)
                webview.set_shown(True)
            # Add startup webpage
            for uri in startup_uris:
                webview = window.container.add_webview_for_uri(
                                                     uri,
                                                     LoadingType.BACKGROUND)
                webview.set_shown(True)
            # Make first webview visible
            if window.container.webviews:
                GLib.idle_add(window.container.set_visible_webview,
                              window.container.webviews[0])
        except Exception as e:
            Logger.error("Application::__restore_state(): %s", e)

    def __on_handle_local_options(self, app, options):
        """
            Handle local options
            @param app as Gio.Application
            @param options as GLib.VariantDict
        """
        if options.contains("version"):
            Logger.info("Eolie %s", self.__version)
            return 0
        return -1

    def __on_command_line(self, app, app_cmd_line):
        """
            Handle command line
            @param app as Gio.Application
            @param options as Gio.ApplicationCommandLine
        """
        self.__externals_count = 0
        args = app_cmd_line.get_arguments()
        options = app_cmd_line.get_options_dict()
        if options.contains("show-tls"):
            self.show_tls = True
        if options.contains("disable-artwork-cache"):
            self.art.disable_cache()

        is_ephemeral = options.contains("private")

        # Only restore state on first run
        if not self.windows:
            self.__init()
            self.__restore_state()

        # Setup at least one window
        if not self.windows or options.contains("new"):
            active_window = self.get_new_window()
        else:
            active_window = self.windows[0]

        if is_ephemeral and len(args) == 1:
            active_window.container.add_webview_for_uri(self.start_page,
                                                        LoadingType.FOREGROUND, True)
        # Open command line args
        for uri in args[1:]:
            # Transform path to uri
            parsed = urlparse(uri)
            if not parsed.scheme:
                if uri.startswith('/'):
                    uri = "file://%s" % uri
                else:
                    uri = "http://%s" % uri
            loading_type = wanted_loading_type(
                len(active_window.container.webviews))
            active_window.container.add_webview_for_uri(uri, loading_type, is_ephemeral)

        # Add default start page
        if not active_window.container.webviews:
            loading_type = wanted_loading_type(
                len(active_window.container.webviews))
            active_window.container.add_webview_for_uri(self.start_page,
                                                        loading_type)
        if self.settings.get_value("debug"):
            WebKit2.WebContext.get_default().get_plugins(None,
                                                         self.__on_get_plugins,
                                                         None)
        Gdk.notify_startup_complete()
        active_window.present()
        return 0

    def __close_window(self, window):
        """
            Close window
        """
        if len(self.get_windows()) > 1:
            state = window.state
            if state is not None:
                self.__state_cache.append(window.state)
                GLib.timeout_add(10000, self.__clean_state_cache, state.wid)
            window.destroy()
        else:
            self.quit(True)

    def __try_closing(self, window, webviews):
        """
            Try closing all webviews
            @param window as Window
            @param webviews as [WebView]
        """
        if webviews:
            webview = webviews.pop(0)
            webview.run_javascript("document.activeElement.tagName;", None,
                                   self.__on_get_active_element,
                                   webviews, window)
        else:
            self.__close_window(window)

    def __on_get_active_element(self, source, result, webviews, window):
        """
            Ask user to close view, if ok, close view
            @param source as GObject.Object
            @param result as Gio.AsyncResult
            @param webviews as [WebView]
            @param window as Window
        """
        def on_response_id(dialog, response_id, window, webviews, self):
            if response_id == Gtk.ResponseType.CLOSE:
                if webviews:
                    self.__try_closing(window, webviews)
                else:
                    self.__close_window(window)
            dialog.destroy()

        def on_close(widget, dialog):
            dialog.response(Gtk.ResponseType.CLOSE)

        def on_cancel(widget, dialog):
            dialog.response(Gtk.ResponseType.CANCEL)

        try:
            data = source.run_javascript_finish(result)
            name = data.get_js_value().to_string()
            if name == "TEXTAREA":
                builder = Gtk.Builder()
                builder.add_from_resource("/org/gnome/Eolie/QuitDialog.ui")
                dialog = builder.get_object("dialog")
                label = builder.get_object("label")
                close = builder.get_object("close")
                cancel = builder.get_object("cancel")
                label.set_text(_("Do you really want to quit Eolie?"))
                dialog.set_transient_for(window)
                dialog.connect("response", on_response_id,
                               window, webviews, self)
                close.connect("clicked", on_close, dialog)
                cancel.connect("clicked", on_cancel, dialog)
                dialog.run()
            else:
                if webviews:
                    self.__try_closing(window, webviews)
                else:
                    self.__close_window(window)
        except Exception as e:
            self.__close_window(window)
            Logger.error("Application::__on_forms_filled(): %s", e)

    def __on_get_plugins(self, source, result, data):
        """
            Print plugins on command line
            @param source as GObject.Object
            @param result as Gio.AsyncResult
            @param data as None
        """
        plugins = source.get_plugins_finish(result)
        for plugin in plugins:
            Logger.info("%s, %s, %s",
                        plugin.get_name(),
                        plugin.get_description(),
                        plugin.get_path())

    def __on_delete_event(self, window, event):
        """
            Exit application
            @param window as Window
            @param event as Gdk.Event
        """
        def on_response_id(dialog, response_id, window):
            if response_id == Gtk.ResponseType.CLOSE:
                self.__try_closing(window, window.container.webviews)
            dialog.destroy()

        def on_close(widget, dialog):
            dialog.response(Gtk.ResponseType.CLOSE)

        def on_cancel(widget, dialog):
            dialog.response(Gtk.ResponseType.CANCEL)
        # Ask for user if needed
        if len(self.get_windows()) == 1 and self.download_manager.active:
            builder = Gtk.Builder()
            builder.add_from_resource("/org/gnome/Eolie/QuitDialog.ui")
            dialog = builder.get_object("dialog")
            label = builder.get_object("label")
            close = builder.get_object("close")
            cancel = builder.get_object("cancel")
            label.set_text(_("Downloads running,"
                             " are you sure you want quit?"))
            dialog.set_transient_for(window)
            dialog.connect("response", on_response_id, window)
            close.connect("clicked", on_close, dialog)
            cancel.connect("clicked", on_cancel, dialog)
            dialog.run()
        else:
            self.__try_closing(window, window.container.webviews)
        return True

    def __on_activate(self, application):
        """
            Call default handler, raise last window
            @param application as Gio.Application
        """
        if not self.windows:
            self.__init()
            self.__restore_state()

        if not self.windows:
            active_window = self.get_new_window()
        else:
            active_window = self.active_window

        if not active_window.container.webviews:
            active_window.container.add_webview_for_uri(self.start_page, LoadingType.FOREGROUND)

        active_window.present()

    def __on_new_window_action(self, action, parameter):
        if not self.windows:
            self.__init()
            self.__restore_state()
            active_window = self.active_window
        else:
            active_window = self.get_new_window()

        if not active_window.container.webviews:
            active_window.container.add_webview_for_uri(self.start_page, LoadingType.FOREGROUND)

        active_window.present()

    def __on_private_page_action(self, action, parameter):
        if not self.windows:
            self.__init()
            self.__restore_state()

        if not self.windows:
            active_window = self.get_new_window()
        else:
            active_window = self.active_window

        active_window.container.add_webview_for_uri(self.start_page, LoadingType.FOREGROUND, True)
        active_window.present()

    def __on_open(self, app, files, _n_files, _hint):
        if not self.windows:
            self.__init()
            self.__restore_state()

        if not self.windows:
            active_window = self.get_new_window()
        else:
            active_window = self.active_window

        for file in files:
            loading_type = wanted_loading_type(
                len(active_window.container.webviews))
            active_window.container.add_webview_for_uri(file.get_uri(), loading_type, False)

        active_window.present()

    def __on_content_blocker_set_filter(self, content_blocker, content_filter):
        """
            Add filter to content manager
            @param content_blocker as ContentBlocker
            @param content_filter as WebKit2.UserContentFilter
        """
        for window in self.windows:
            for webview in window.container.webviews:
                content_manager = webview.get_user_content_manager()
                content_manager.add_filter(content_filter)

    def __on_content_blocker_unset_filter(self, content_blocker,
                                          content_filter):
        """
            Remove filter from content manager
            @param content_blocker as ContentBlocker
            @param content_filter as WebKit2.UserContentFilter
        """
        for window in self.windows:
            for webview in window.container.webviews:
                content_manager = webview.get_user_content_manager()
                content_manager.remove_filter(content_filter)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib

from eolie.sites_manager import SitesManager
from eolie.define import App


class SidebarContainer:
    """
        Sidebar management for container
    """

    def __init__(self):
        """
            Init container
        """
        self.__sites_manager = SitesManager(self._window)
        self.set_orientation(Gtk.Orientation.HORIZONTAL)
        position = App().settings.get_value("sidebar-position").get_int32()
        self.set_position(position)
        self.__sites_manager.show_labels(position > 60)
        self.connect("notify::position", self.__on_paned_notify_position)
        self.pack1(self.__sites_manager, False, False)
        if App().settings.get_value("show-sidebar"):
            self.__sites_manager.show()
        App().settings.connect("changed::show-sidebar",
                               self.__on_show_sidebar_changed)

    @property
    def sites_manager(self):
        """
            Get sites manager
            @return SitesManager
        """
        return self.__sites_manager

#######################
# PRIVATE             #
#######################
    def __on_paned_notify_position(self, paned, ignore):
        """
            Save position
            @param paned as Gtk.Paned
            @param ignore as GParamInt
        """
        position = paned.get_position()
        App().settings.set_value("sidebar-position",
                                 GLib.Variant("i", position))
        self.__sites_manager.show_labels(position > 60)

    def __on_show_sidebar_changed(self, settings, key):
        """
            Show/hide panel
            @param settings as Gio.Settings
            @param key as str
        """
        if App().settings.get_value("show-sidebar"):
            self.__sites_manager.show()
        else:
            self.__sites_manager.hide()



# Copyright (c) 2014-2017 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


class LinkedList:
    """
        Double linked list
    """

    def __init__(self, value=None, next=None, prev=None):
        """
            Init list
            @param value as int
        """
        self.__value = value
        self.__next = next
        self.__prev = prev

    def has_next(self):
        """
            True if list has next
            @return has next as bool
        """
        return self.__next is not None

    def has_prev(self):
        """
            True if list has prev
            @return has next as bool
        """
        return self.__prev is not None

    def set_next(self, next):
        """
            Set next
            @param next as linked list
        """
        self.__next = next

    def set_prev(self, prev):
        """
            Set prev
            @param prev as linked list
        """
        self.__prev = prev

    @property
    def next(self):
        """
            Return next
            @return next as LinkedList
        """
        return self.__next

    @property
    def prev(self):
        """
            Return prev
            @return prev as LinkedList
        """
        return self.__prev

    @property
    def value(self):
        """
            Get value
            @return value as int
        """
        return self.__value



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GObject

from gettext import gettext as _
from urllib.parse import urlparse

from eolie.logger import Logger
from eolie.define import App
from eolie.utils import emit_signal
from eolie.helper_gestures import GesturesHelper


class ScriptRow(Gtk.ListBoxRow, GesturesHelper):
    """
        Script row (Allowing to select a script for uri)
    """

    __gsignals__ = {
        "activated": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, uri, toggled):
        """
            Init row
            @param uri as str
            @param toggled as bool
        """
        Gtk.ListBoxRow.__init__(self)
        self.__uri = uri
        grid = Gtk.Grid()
        grid.set_column_spacing(5)
        grid.show()
        label = Gtk.Label.new(uri)
        label.set_hexpand(True)
        label.set_property("halign", Gtk.Align.START)
        label.set_tooltip_text(uri)
        label.show()
        self.__check = Gtk.CheckButton()
        self.__check.set_active(toggled)
        self.__check.connect("toggled", self.__on_check_toggled)
        self.__check.show()
        grid.add(self.__check)
        grid.add(label)
        eventbox = Gtk.EventBox.new()
        eventbox.show()
        eventbox.add(grid)
        self.add(eventbox)
        GesturesHelper.__init__(self, eventbox)

    def set_active(self, active):
        """
            Set row active
            @param active as bool
        """
        self.__check.set_active(active)

    @property
    def uri(self):
        """
            Get uri
            @return str
        """
        return self.__uri

    @property
    def is_active(self):
        """
            Get Check button
            @return bool
        """
        return self.__check.get_active()

#######################
# PROTECTED           #
#######################
    def _on_primary_press_gesture(self, x, y, event):
        """
            Toggle check button
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        toggled = not self.__check.get_active()
        self.__check.set_active(toggled)

#######################
# PRIVATE             #
#######################
    def __on_check_toggled(self, button):
        """
            Emit activated signal
            @param button as Gtk.CheckButton
        """
        emit_signal(self, "activated")


class ScriptsMenu(Gtk.Grid):
    """
        Menu allowing user to enable scripts for URI
    """

    __ALLOW_ALL = _("Allow all scripts")

    def __init__(self, uri, window):
        """
            Init widget
            @param uri as str
            @param window as Window
        """
        Gtk.Grid.__init__(self)
        self.set_column_spacing(10)
        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.__uri = uri
        self.__window = window
        self.__back_button = Gtk.ModelButton.new()
        self.__back_button.show()
        self.__back_button.set_property("menu_name", "main")
        self.__back_button.set_property("inverted", True)
        self.__back_button.set_property("centered", True)
        self.__back_button.set_property("text", _("Scripts"))
        self.__scrolled = Gtk.ScrolledWindow.new()
        self.__scrolled.show()
        self.__scrolled.set_size_request(-1, 300)
        self.__scrolled.set_shadow_type(Gtk.ShadowType.NONE)
        self.__listbox = Gtk.ListBox.new()
        self.__listbox.show()
        self.__listbox.get_style_context().add_class("menu-listbox")
        self.__listbox.set_selection_mode(Gtk.SelectionMode.NONE)
        self.__listbox.connect("map", self.__on_map)
        self.set_property("margin", 5)
        self.__scrolled.add(self.__listbox)
        self.add(self.__back_button)
        self.add(self.__scrolled)

#######################
# PRIVATE             #
#######################
    def __is_script_active(self, script, netloc, internal):
        """
            True if script active for netloc
            @param script as str
            @param netloc as str
            @param internal as bool
            @return bool
        """
        content_blocker = App().get_content_blocker("block-scripts")
        if script == ".*":
            uri = ".*"
        else:
            script_parsed = urlparse(script)
            uri = "%s%s.*" % (script_parsed.netloc, script_parsed.path)
        result = content_blocker.exceptions.is_domain_exception(
            netloc, uri, internal)
        return result

    def __on_map(self, listbox):
        """
            Populate languages
            @param listbox as Gtk.ListBox
        """
        if listbox.get_children():
            return
        webview = self.__window.container.webview
        webview.run_javascript_from_gresource(
                "/org/gnome/Eolie/javascript/GetScripts.js", None,
                self.__on_get_scripts)

    def __on_get_scripts(self, source, result):
        """
            Add scripts to menu
            @param source as GObject.Object
            @param result as Gio.AsyncResult
        """
        scripts = []
        try:
            data = source.run_javascript_from_gresource_finish(result)
            scripts = data.get_js_value().to_string().split("@@@")
        except Exception as e:
            Logger.warning("ScriptsMenu::__on_get_scripts(): %s", e)
        parsed = urlparse(self.__window.container.webview.uri)
        allow_all_active = self.__is_script_active('.*', parsed.netloc, False)
        row = ScriptRow(self.__ALLOW_ALL, allow_all_active)
        row.connect("activated", self.__on_row_activated, False)
        row.show()
        self.__listbox.add(row)
        for script in scripts:
            if not script:
                continue
            internal = script.find(parsed.netloc) != -1
            row = ScriptRow(script,
                            self.__is_script_active(script,
                                                    parsed.netloc,
                                                    internal))
            row.connect("activated", self.__on_row_activated, internal)
            if allow_all_active:
                row.set_sensitive(False)
            row.show()
            self.__listbox.add(row)

    def __on_row_activated(self, row, internal):
        """
            Update exceptions
            @param row as ScriptRow
            @param internal as bool
        """
        parsed = urlparse(self.__window.container.webview.uri)
        content_blocker = App().get_content_blocker("block-scripts")
        if row.uri == self.__ALLOW_ALL:
            uri = ".*"
            for _row in self.__listbox.get_children()[1:]:
                _row.set_active(False)
                _row.set_sensitive(not row.is_active)
        else:
            script_parsed = urlparse(row.uri)
            uri = "%s%s.*" % (script_parsed.netloc, script_parsed.path)
        if row.is_active:
            content_blocker.exceptions.add_domain_exception(
                parsed.netloc, uri, internal)
        else:
            content_blocker.exceptions.remove_domain_exception(
                parsed.netloc, uri, internal)
        content_blocker.exceptions.save()
        content_blocker.update()
        self.__window.container.webview.reload()



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gio

from gettext import gettext as _

from eolie.define import App, Type
from eolie.popover_uri_item import Item
from eolie.popover_uri_row import Row
from eolie.popover_uri_input import Input


class UriPopoverContent:
    """
        Content handler for UriPopover
    """

    def __init__(self):
        """
            Init handler
        """
        self._input = None
        self._bookmarks_model = Gio.ListStore()
        self._history_model = Gio.ListStore()

    def search_value(self, value, cancellable):
        """
           Search value
           @param value as str
           @param cancellable as Gio.Cancellable
        """
        self._task_helper.run(self.__search_value, value, cancellable)

#######################
# PROTECTED           #
#######################
    def _add_bookmarks(self, bookmarks):
        """
            Add bookmarks to model
            @param [(bookmark_id, title, uri)] as [(int, str, str)]
        """
        if bookmarks:
            (bookmark_id, uri, title) = bookmarks.pop(0)
            item = Item()
            item.set_property("id", bookmark_id)
            item.set_property("type", Type.BOOKMARK)
            item.set_property("title", title)
            item.set_property("uri", uri)
            self._bookmarks_model.append(item)
            GLib.idle_add(self._add_bookmarks, bookmarks)

    def _add_tags(self, tags, select, position=0):
        """
            Add tags to model
            @param [(tag_id, title)] as [(int, str)]
            @param select as int
        """
        if tags:
            (tag_id, title) = tags.pop(0)
            item = Item()
            item.set_property("id", tag_id)
            item.set_property("type", Type.TAG)
            item.set_property("title", title)
            child = Row(item, self._window)
            child.connect("activate", self.__on_row_activated)
            child.connect("moved", self.__on_row_moved)
            child.show()
            self._tags_box.add(child)
            GLib.idle_add(self._add_tags, tags, select)
        else:
            if select is None:
                select = Type.POPULARS
            # Search for previous current row
            for row in self._tags_box.get_children():
                if row.item.get_property("id") == select:
                    self._tags_box.select_row(row)
                    break
            self._set_bookmarks(select)

    def _add_history_items(self, items, date):
        """
            Add history items to model
            @param [(history_id, title, uri, atime)]  as [(int, str, str, int)]
            @param date (jj, mm, aaaa) as (int, int, int)
        """
        if items:
            if date != self._calendar.get_date():
                return
            (history_id, title, uri, atime) = items.pop(0)
            item = Item()
            item.set_property("id", history_id)
            item.set_property("type", Type.HISTORY)
            item.set_property("title", title)
            item.set_property("uri", uri)
            item.set_property("atime", atime)
            self._history_model.append(item)
            GLib.idle_add(self._add_history_items, items, date)

    def _set_bookmarks(self, tag_id):
        """
            Set bookmarks for tag id
            @param tag id as int
        """
        self._bookmarks_model.remove_all()
        self._remove_button.hide()
        if tag_id == Type.POPULARS:
            items = App().bookmarks.get_populars(50)
        elif tag_id == Type.RECENTS:
            items = App().bookmarks.get_recents()
        elif tag_id == Type.UNCLASSIFIED:
            items = App().bookmarks.get_unclassified()
        else:
            items = App().bookmarks.get_bookmarks(tag_id)
        self._bookmarks_count.set_text(_("%s bookmarks") % len(items))
        GLib.idle_add(self._add_bookmarks, items)

    def _get_current_box(self):
        """
            Get current box
            @return Gtk.ListBox
        """
        name = self._stack.get_visible_child_name()
        box = None
        if name == "bookmarks":
            box = self._bookmarks_box
        elif name == "history":
            box = self._history_box
        elif name == "search":
            box = self._search_box
        return box

    def _get_current_input_box(self):
        """
            Get current box with input
            @return Gtk.ListBox
        """
        box = None
        if self._input == Input.SEARCH:
            box = self._search_box
        elif self._input == Input.TAGS:
            box = self._tags_box
        elif self._input == Input.BOOKMARKS:
            box = self._bookmarks_box
        return box

#######################
# PRIVATE             #
#######################
    def __search_value(self, value, cancellable):
        """
            Search for value in DB
            @param value as str
            @param cancellable as Gio.Cancellable
        """
        if value == "":
            result = App().history.get_populars(25)
        else:
            result = App().bookmarks.search(value, 15)
            result += App().history.search(value, 15)
        GLib.idle_add(self.__add_searches, result, cancellable)

    def __add_searches(self, result, cancellable):
        """
            Add searches to model
            @param result as [(str, str, int)]
            @param cancellable as Gio.Cancellable
        """
        if cancellable.is_cancelled():
            return
        if result:
            (rowid, title, uri) = result.pop(0)
            item = Item()
            item.set_property("id", rowid)
            item.set_property("type", Type.SEARCH)
            item.set_property("title", title)
            item.set_property("uri", uri)
            child = Row(item, self._window)
            child.show()
            self._search_box.add(child)
            GLib.idle_add(self.__add_searches, result, cancellable)

    def __on_row_activated(self, row):
        """
            Select row
            @param row as Row
        """
        item_id = row.item.get_property("id")
        self._set_bookmarks(item_id)

    def __on_row_moved(self, row, items):
        """
            Move bookmark from current selected tag to tag
            @param row as Row
            @param items [(bookmark_id, tag_id)] as [(int, int)]
            @param tag id as int
        """
        tag_row = self._tags_box.get_selected_row()
        current_tag_id = tag_row.item.get_property("id")
        for item in items:
            if current_tag_id >= 0:
                App().bookmarks.del_tag_from(current_tag_id, item[0])
            App().bookmarks.add_tag_to(item[1], item[0])
        self.__on_row_activated(tag_row)
        App().bookmarks.clean_tags()
        if App().sync_worker is not None:
            App().sync_worker.push_bookmark(item[0])



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GObject, Pango, GLib

from eolie.helper_passwords import PasswordsHelper
from eolie.define import App, MARGIN_SMALL
from eolie.utils import emit_signal


class Item(GObject.GObject):
    username = GObject.Property(type=str,
                                default="")
    uri = GObject.Property(type=str,
                           default="")
    uuid = GObject.Property(type=str,
                            default="")

    def __init__(self):
        GObject.GObject.__init__(self)


class Row(Gtk.ListBoxRow):
    """
        A cookie row
    """

    def __init__(self, item, helper):
        """
            Init row
            @param item as Item
            @param helper as PasswordsHelper
        """
        Gtk.ListBoxRow.__init__(self)
        self.__item = item
        self.__helper = helper
        self.get_style_context().add_class("row")
        uri = Gtk.Label.new()
        uri.set_markup("%s <span alpha='40000'>%s</span>" % (
            GLib.markup_escape_text(item.get_property("uri")),
            GLib.markup_escape_text(item.get_property("username"))))
        uri.set_ellipsize(Pango.EllipsizeMode.END)
        uri.set_property("margin", MARGIN_SMALL)
        uri.set_property("halign", Gtk.Align.START)
        uri.set_hexpand(True)
        uri.set_property("has-tooltip", True)
        uri.connect("query-tooltip", self.__on_query_tooltip)
        uri.show()
        self.add(uri)

    def delete(self):
        """
            Delete password
        """
        self.hide()
        uuid = self.__item.get_property("uuid")
        self.__helper.clear(uuid)
        if App().sync_worker is not None:
            App().sync_worker.remove_from_passwords(uuid)

    @property
    def item(self):
        """
            Get item for row
            @return Item
        """
        return self.__item

#######################
# PRIVATE             #
#######################
    def __on_delete_clicked(self, button):
        """
            Delete item from cookies
            @param button as Gtk.Button
        """
        self.delete()

    def __on_query_tooltip(self, widget, x, y, keyboard, tooltip):
        """
            Show tooltip if needed
            @param widget as Gtk.Widget
            @param x as int
            @param y as int
            @param keyboard as bool
            @param tooltip as Gtk.Tooltip
        """
        text = ""
        layout = widget.get_layout()
        label = widget.get_text()
        if layout.is_ellipsized():
            text = "%s" % (GLib.markup_escape_text(label))
        widget.set_tooltip_markup(text)


class CredentialsDialog(Gtk.Bin):
    """
        Show saved credentials
    """

    __gsignals__ = {
        "destroy-me": (GObject.SignalFlags.RUN_FIRST, None, ())
    }

    def __init__(self):
        """
            Init dialog
        """
        Gtk.Bin.__init__(self)
        self.__filter = ""
        self.__helper = PasswordsHelper()
        builder = Gtk.Builder()
        builder.add_from_resource('/org/gnome/Eolie/DialogCredentials.ui')
        builder.connect_signals(self)
        self.__search_bar = builder.get_object("search_bar")
        self.__remove_button = builder.get_object("remove_button")
        self.__listbox = builder.get_object("listbox")
        self.__listbox.set_filter_func(self.__filter_func)
        self.__listbox.set_sort_func(self.__sort_func)
        self.add(builder.get_object('widget'))
        self.__helper.get_all(self.__add_password)

#######################
# PROTECTED           #
#######################
    def _on_back_clicked(self, button):
        """
            Ask to be destroyed
            @param button as Gtk.Button
        """
        emit_signal(self, "destroy-me")

    def _on_search_changed(self, entry):
        """
            Update filter
            @param entry as Gtk.Entry
        """
        self.__filter = entry.get_text()
        self.__listbox.invalidate_filter()

    def _on_remove_clicked(self, button):
        """
            Remove all passwords
            @param button as Gtk.Button
        """
        for row in self.__listbox.get_selected_rows():
            row.delete()

    def _on_row_selected(self, listbox, row):
        """
            Update clear button state
            @param listbox as Gtk.ListBox
            @param row as Gtk.ListBoxRow
        """
        self.__remove_button.set_sensitive(
            len(listbox.get_selected_rows()) != 0)

    def _on_search_toggled(self, button):
        """
            Show entry
            @param button as Gtk.Button
        """
        self.__search_bar.set_search_mode(button.get_active())

#######################
# PRIVATE             #
#######################
    def __filter_func(self, row):
        """
            Filter rows
            @param row as Row
        """
        return self.__filter in row.item.get_property("uri")

    def __sort_func(self, row1, row2):
        """
            Sort rows
            @param row1 as Row
            @param row2 as Row
        """
        return row2.item.get_property("username") <\
            row1.item.get_property("username")

    def __add_password(self, attributes, password, uri, index, count):
        """
            Add password to model
            @param attributes as {}
            @param password as str
            @param uri as None
            @param index as int
            @param count as int
        """
        if attributes is None:
            return
        try:
            item = Item()
            item.set_property("username", attributes["login"])
            item.set_property("uri", attributes["formSubmitURL"])
            item.set_property("uuid", attributes["uuid"])
            child = Row(item, self.__helper)
            child.show()
            self.__listbox.add(child)
        except:
            # Here because firsts Eolie releases do not
            # provide formSubmitURL
            # FIXME Remove later
            pass
